/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                 */
/* file: aPacketInternal.c                                         */
/*                                                                 */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                 */
/* description: Implementation of a platform-independent BrainStem */
/*		communication layer.    			   */
/*                                                                 */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                 */
/* Copyright  1997-2000. Acroname Inc.                            */
/*                                                                 */
/* This software is the property of Acroname Inc.  Any             */
/* distribution, sale, transmission, or re-use of this code is     */
/* strictly forbidden except with permission from Acroname Inc.    */
/*                                                                 */
/* To the full extent allowed by law, Acroname Inc. also excludes  */
/* for itself and its suppliers any liability, wheither based in   */
/* contract or tort (including negligence), for direct,            */
/* incidental, consequential, indirect, special, or punitive       */
/* damages of any kind, or for loss of revenue or profits, loss of */
/* business, loss of information or data, or other financial loss  */
/* arising out of or in connection with this software, even if     */
/* Acroname Inc. has been advised of the possibility of such       */
/* damages.                                                        */
/*                                                                 */
/* Acroname Inc.                                                   */
/* www.acroname.com                                                */
/* 720-564-0373                                                    */
/*                                                                 */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "aStemInternal.h"
#include "aPacketInternal.h"


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aPacket_Init
 */

void aPacket_Init(aPacket* pPacket,
		  unsigned char initflags)
{
  if (!pPacket)
    return;

  aBZero(pPacket, sizeof(aPacket));

  pPacket->check = aPACKETCHECK;
  
  pPacket->status = initflags; 

} /* end of aPacket_Init */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aPacketInternal_Create
 */

aErr aPacketInternal_Create(void* vpStem,
			    unsigned char address,
			    unsigned char nLength,
			    char* data,
			    aPacket** ppPacket)
{
  aStem* pStem = (aStem*)vpStem;
  aErr stemErr = aErrNone;

  aAssert(pStem);
  aAssert(pStem->packetPoolRef);

  if ((data == NULL) || (ppPacket == NULL))
    stemErr = aErrParam;

  /* packet is limited in size */
  if ((stemErr == aErrNone) && (nLength > aSTEMMAXPACKETBYTES))
    stemErr = aErrSize;

  /* address can only be even */
  if ((stemErr == aErrNone) && 
      ((address & 0x01) && (address != aSTEM_MAGICADDR)))
    stemErr = aErrParam;

  /* build the packet */
  if (stemErr == aErrNone)
    aMemPool_Alloc(pStem->ioRef, pStem->packetPoolRef, 
    		   (void*)ppPacket, &stemErr);
  
  if (stemErr == aErrNone) {
    aPacket_Init(*ppPacket, 0);
    (*ppPacket)->address = address;
    (*ppPacket)->dataSize = nLength;
    aMemCopy((*ppPacket)->data, data, nLength);
  }

  return(stemErr);

} /* end of aPacketInternal_Create */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aPacket_Size
 */

int aPacket_Size(aPacket* pPacket)
{
  if (!pPacket)
    return 0;

  return(pPacket->dataSize);
  
} /* end of aPacket_Size */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aPacket_BytesSoFar
 */

int aPacket_BytesSoFar(aPacket* pPacket)
{
  if (!pPacket)
    return 0;

  return(pPacket->curSize);

} /* end of aPacket_BytesSoFar routine */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aPacket_Complete
 */

aBool aPacket_Complete(aPacket* pPacket)
{
  if (!pPacket)
    return aFalse;

  if (!(pPacket->status & aPacketHasSize))
    return aFalse;

  if (pPacket->curSize == pPacket->dataSize + 2)
    return aTrue;

  return aFalse;
  
} /* end of aPacket_Complete routine */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aPacket_AddByte
 */

aErr aPacket_AddByte(aPacket* pPacket, 
		     unsigned char byte)
{
  if (!pPacket)
    return aErrParam;

  pPacket->curSize++;

  /* account for cmd and address */
  if (pPacket->curSize > aSTEMMAXPACKETBYTES + 2)
    return aErrUnknown;

  if (!(pPacket->status & aPacketHasAddr)) {
    pPacket->status |= aPacketHasAddr;
    pPacket->address = byte;
    return aErrNone;
  }

  if (!(pPacket->status & aPacketHasSize)) {
    pPacket->status |= aPacketHasSize;
    pPacket->dataSize = byte;
    if (pPacket->dataSize > aSTEMMAXPACKETBYTES)
      return aErrUnknown;
    return aErrNone;
  }

  pPacket->data[pPacket->curSize - 3] = (char)byte;
  
  return aErrNone;

} /* end of aPacket_AddByte routine */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aPacketInternal_Format
 */

aErr aPacketInternal_Format(const aPacket* pPacket,
			    char* pBuffer,
			    unsigned short nMaxLength)
{
  aErr stemErr = aErrNone;
  char* p = pBuffer;
  unsigned short len = 0;
  int i;
  short  visLen = (short)(nMaxLength - 1);
  char hexlookup[16] = {'0', '1', '2', '3', 
			'4', '5', '6', '7',
			'8', '9', 'A', 'B', 
			'C', 'D', 'E', 'F'};

  aVALIDPACKET(pPacket);

  if (stemErr == aErrNone) {
    if (len++ >= visLen) goto end;
    *p++ = (char)((pPacket->status & aPacketFromStem) ? '<' : '>');
  
    if (len++ >= visLen) goto end;
    *p++ = ' ';

    if (len++ >= visLen) goto end;
    *p++ = hexlookup[(pPacket->address >> 4) & 0x0F];

    if (len++ >= visLen) goto end;
    *p++ = hexlookup[pPacket->address & 0x0F];

    if (len++ >= visLen) goto end;
    *p++ = ':';

    for (i = 0; (len++ < visLen) && (i < pPacket->dataSize); i++) {
      *p++ = hexlookup[(pPacket->data[i] >> 4) & 0x0F];
      if (len++ >= visLen) goto end;
      *p++ = hexlookup[pPacket->data[i] & 0x0F];
      if (i < pPacket->dataSize - 1) {
        if (len++ >= visLen) goto end;
        *p++ = ',';
      }
    }
  
  end:
    *p = 0; /* null terminate */
  }

  return(stemErr);

} /* end of aPacketInternal_Format routine */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aPacketInternal_Destroy
 */

aErr aPacketInternal_Destroy(void* vpStem,
	         	     aPacket* pPacket)
{
  aStem* pStem = (aStem*)vpStem;
  aErr stemErr = aErrNone;

  aVALIDSTEM(pStem);
  aVALIDPACKET(pPacket);

  if (stemErr == aErrNone) {
    aAssert(pStem->packetPoolRef);

    /* invalidate the packet */
    pPacket->check = 0;
    aMemPool_Free(pStem->ioRef, pStem->packetPoolRef, 
    		  pPacket, &stemErr);
  }

  return stemErr;

} /* aPacketInternal_Destroy */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aPacketInternal_Destroy
 */

aErr aPacketInternal_GetData(void* vpStem,
			     const aPacketRef packetRef,
			     unsigned char* pAddress,
			     unsigned char* pLen,
			     char* data)
{
  aErr stemErr = aErrNone;
  aStem* pStem = (aStem*)vpStem;
  aPacket* pPacket = (aPacket*)packetRef;

  /* basic parameter checking */
  if ((pStem == NULL) || 
      (pPacket == NULL) || 
      (data == NULL) ||
      (pLen == NULL) ||
      (pAddress == NULL))
    stemErr = aErrParam;

  if (stemErr == aErrNone) {
    aAssert(pPacket->dataSize <= aSTEMMAXPACKETBYTES);
    aAssert(!(pPacket->address & 0x01));
    *pAddress = pPacket->address;
    *pLen = pPacket->dataSize;
    aMemCopy(data, pPacket->data, pPacket->dataSize);
  }
  
  return stemErr;

} /* aPacketInternal_GetData */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aPacketInternal_WriteToStream
 */
 
aErr aPacketInternal_WriteToStream(aPacket* pPacket,
				   aStreamRef stream)
{
  aErr stemErr = aErrNone;
  unsigned char buffer[aSTEMMAXPACKETBYTES + 2]; /* data + length + address */

  /* accumulate the data into a single buffer */
  buffer[0] = pPacket->address;  /* address */ 
  buffer[1] = pPacket->dataSize; /* length */
  aMemCopy(&buffer[2], pPacket->data, pPacket->dataSize); /* data */

  /* dump to the stream */ 
  if (stemErr == aErrNone)
    aStream_Write(aStreamLibRef(stream),
  		  stream,
  		  (char*)buffer,
  		  (unsigned long)(pPacket->dataSize + 2), 
  		  &stemErr);

  return stemErr;
 
} /* aPacketInternal_WriteToStream */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aPacketList_Create
 */

aErr aPacketList_Create(aMemPoolRef packetPoolRef,
			aPacketList** ppPacketList)
{
  aErr plErr = aErrNone;
  aPacketList* pPacketList;

  aAssert(packetPoolRef);
  aAssert(ppPacketList);

  if (plErr == aErrNone) {
    pPacketList = (aPacketList*)aMemAlloc(sizeof(aPacketList));
    if (pPacketList == NULL) {
      plErr = aErrMemory;
    } else {
      aBZero(pPacketList, sizeof(aPacketList));
      pPacketList->packetPoolRef = packetPoolRef;
      *ppPacketList = pPacketList;
    }
  }

  return plErr;

} /* aPacketList_Create */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aPacketList_AddCopy
 *
 * adds a packet to the tail of the list
 */

aErr aPacketList_AddCopy(void* vpStem,
			 aPacketList* pPacketList,
		         aPacket* pPacket)
{
  aStem* pStem = (aStem*)vpStem;
  aErr stemErr = aErrNone;
  aPacket* pCopy;

  aAssert(pPacketList);

  /* create the copy */
  if (stemErr == aErrNone)
    aMemPool_Alloc(pStem->ioRef, pPacketList->packetPoolRef, 
		   (void*)&pCopy, &stemErr);
 
  /* copy it from the passed-in Packet */
  if (stemErr == aErrNone) 
    aMemCopy(pCopy, pPacket, sizeof(aPacket));

  /* insert it at the end of the list */
  if (stemErr == aErrNone) {
    pPacketList->nSize++;
    pCopy->pNext = NULL;
    if (pPacketList->pHead == NULL) {
      pPacketList->pHead = pCopy;
    } else {
      pPacketList->pTail->pNext = pCopy;
    }
    pPacketList->pTail = pCopy;
  }

  return stemErr;  

} /* aPacketList_AddCopy */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aPacketList_GetFirst
 * 
 * Starts looking at the front of the list for a packet.  This 
 * yields the oldest packet first. If a filter is passed in, it 
 * is used, otherwise, the first available packet is returned.
 */

aErr aPacketList_GetFirst(aPacketList* pPacketList,
			  aPacketFilter filterProc,
			  void* filterRef,
		          aPacket** ppPacket)
{
  aErr stemErr = aErrNone;

  aAssert(pPacketList);
  aAssert(ppPacket);

  if (filterProc == NULL) {
  
    /* with no filter proc, just run down the list and look for 
     * a list head to find the first packet
     */
    if (pPacketList->pHead == NULL)
      stemErr = aErrNotFound;
    else {
      pPacketList->nSize--;
      *ppPacket = pPacketList->pHead;
      pPacketList->pHead = pPacketList->pHead->pNext;
    }
  } else {

    /* with a filter proc, search the list to try to find 
     * the first packet the filter returns true on.
     */
    aPacket* pPacket = pPacketList->pHead;
    aPacket* pPrev = NULL;
    while ((pPacket != NULL) 
           && (filterProc(pPacket->address,
    			  pPacket->dataSize,
    			  pPacket->data,
    			  filterRef) == aFalse)
    	   && (!aPACKET_IS_HB(pPacket))) {
      pPrev = pPacket;
      pPacket = pPacket->pNext;
    }
    if (pPacket == NULL)
      stemErr = aErrNotFound;
    else {
      /* record the find and remove it from the list */
      pPacketList->nSize--;
      *ppPacket = pPacket;
      if (pPrev == NULL)
        pPacketList->pHead = pPacket->pNext;
      else {
        pPrev->pNext = pPacket->pNext;
        if (pPacketList->pTail == pPacket)
          pPacketList->pTail = pPrev;
      }
    }
  }

  return stemErr;

} /* aPacketList_GetFirst */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aPacketList_Destroy
 */

aErr aPacketList_Destroy(void* vpStem,
			 aPacketList* pPacketList)
{
  aStem* pStem = (aStem*)vpStem;
  aErr stemErr = aErrNone;
  aPacket* pPacket;
  
  aAssert(pPacketList);
  aAssert(pPacketList->packetPoolRef);
  
  /* remove all the list elements */
  while((stemErr == aErrNone) && (pPacketList->pHead != NULL)) {
    pPacket = pPacketList->pHead;
    pPacketList->pHead = pPacket->pNext;
    aMemPool_Free(pStem->ioRef, pPacketList->packetPoolRef, 
    		  pPacket, &stemErr);
  }
  
  /* blow away the list itself */
  if (stemErr == aErrNone)
    aMemFree((aMemPtr)pPacketList);

  return stemErr;

} /* aPacketList_Destroy */
