/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                 */
/* file: aSettingFileInternal.c                                    */
/*                                                                 */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                 */
/* Copyright  1997-2001. Acroname Inc.                            */
/*                                                                 */
/* This software is the property of Acroname Inc.  Any             */
/* distribution, sale, transmission, or re-use of this code is     */
/* strictly forbidden except with permission from Acroname Inc.    */
/*                                                                 */
/* To the full extent allowed by law, Acroname Inc. also excludes  */
/* for itself and its suppliers any liability, wheither based in   */
/* contract or tort (including negligence), for direct,            */
/* incidental, consequential, indirect, special, or punitive       */
/* damages of any kind, or for loss of revenue or profits, loss of */
/* business, loss of information or data, or other financial loss  */
/* arising out of or in connection with this software, even if     */
/* Acroname Inc. has been advised of the possibility of such       */
/* damages.                                                        */
/*                                                                 */
/* Acroname Inc.                                                   */
/* www.acroname.com                                                */
/* 720-564-0373                                                    */
/*                                                                 */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "aOSDefs.h"
#include "aIOInternal.h"
#include "aSymbolTableInternal.h"
#include "aSettingFileInternal.h"
#include "aUtil.h"


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * local prototypes
 */

static void sSettingFile_ReadLine(aSettingFile* pSettings);
static aErr sSettingFile_Read(aSettingFile* pSettings,
			      const char* pFileName);
static aErr sSettingDeleteProc(void* pData, void* ref);


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * sSettingDeleteProc
 */

aErr sSettingDeleteProc(void* pData, void* ref)
{
  if (pData == NULL)
    return aErrParam;

  aMemFree(pData);

  return aErrNone;

} /* sSettingDeleteProc */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * sSettingFile_ReadLine
 */

void sSettingFile_ReadLine(aSettingFile* pSettings)
{
  char* p = pSettings->readBuf;
  char key[100];
  char* k = key;
  char value[100];
  char* v = value;
  char* pData;
    
  /* skip over initial whitespace */
  while (*p && ((*p == ' ') || (*p == '\t'))) p++;

  /* ignore comments starting with "#" */  
  if (*p == '#')
    return;

  /* move up to the equal sign */
  while (*p && ((*p != ' ') && (*p != '\t') && (*p != '='))) {
    *k++ = *p++;
  }

  /* null terminate the key */
  *k = 0;
  
  if (!*p)
    return;

  /* skip over the equal and any whitespace */
  while (*p && ((*p == ' ') || (*p == '\t') || (*p == '='))) p++;

  if (!*p)
    return;

  /* move up to the end of line */
  while (*p)
    *v++ = *p++;
  
  /* null terminate the value */
  *v = 0;
  
  pData = (char*)aMemAlloc((aMemSize)(aStringLen(value) + 1));
  if (!pData)
    return;
  
  aStringCopy(pData, value);

  aSymbolTableInternal_Insert(pSettings->pSettingTable, 
  		      	      key, pData, sSettingDeleteProc, NULL);

} /* sSettingFile_ReadLine */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * sSettingFile_Read
 */

aErr sSettingFile_Read(aSettingFile* pSettings,
		       const char* pFileName)
{
  aErr settingErr = aErrNone;
  aErr ioErr = aErrNone;
  aStreamRef file = NULL;

  /* open the file stream */
  if (settingErr == aErrNone) {
    if (aStream_CreateFileInput(pSettings->ioLib, 
    				pFileName, aFileAreaBinary,  
    		   		&file, &ioErr))
    settingErr = aErrNotFound;
  }

  /* read in the file */  
  if (settingErr == aErrNone) {
    do {
      aStream_ReadLine(pSettings->ioLib, file, pSettings->readBuf, 
      		       pSettings->nMaxSettingLen, &ioErr);
      if (ioErr == aErrNone) {
        /* we now have a line, parse it and store in the 
         * symbol table */
        sSettingFile_ReadLine(pSettings);
      }
    } while (ioErr == aErrNone);
  }

  /* close the file stream */
  if (file != NULL)
    aStream_Destroy(pSettings->ioLib, file, NULL);

  /* just fall back on defaults if file not found */
  if (settingErr == aErrNotFound)
    settingErr = aErrNone;

  return settingErr;

} /* sSettingFile_Read */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aSettingFileInternal_Create
 */

aErr aSettingFileInternal_Create(aIOLib ioRef,
				 const unsigned int nMaxSettingLen,
			 	 const char* pFileName,
			 	 aSettingFile** ppSettingFile)
{
  aErr settingErr = aErrNone;
  aSettingFile* pSettingFile = NULL;

  if ((ppSettingFile == NULL) || (pFileName == NULL))
    settingErr = aErrParam;

  if (settingErr == aErrNone) {
    pSettingFile = (aSettingFile*)aMemAlloc(sizeof(aSettingFile));
    if (pSettingFile == NULL)
      settingErr = aErrMemory;
  }

  if (settingErr == aErrNone) {
    aBZero(pSettingFile, sizeof(aSettingFile));
    pSettingFile->check = aSETTINGFILECHECK;
    pSettingFile->ioLib = ioRef;
    pSettingFile->nMaxSettingLen = nMaxSettingLen;
    settingErr = aMemPoolInternal_Create((aMemSize)nMaxSettingLen, 8, 
      		    			 &pSettingFile->pSettingPool);
  }

  if (settingErr == aErrNone)
    settingErr = aSymbolTableInternal_Create(&pSettingFile->pSettingTable); 

  if (settingErr == aErrNone)
    settingErr = sSettingFile_Read(pSettingFile, pFileName);

  if (settingErr == aErrNone)
    *ppSettingFile = pSettingFile;
  else if (pSettingFile != NULL)
    aSettingFileInternal_Destroy(pSettingFile);

  return settingErr;

} /* aSettingFileInternal_Create */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aSettingFileInternal_GetInt
 */

aErr aSettingFileInternal_GetInt(aSettingFile* pSettingFile,
			 	 const char* key, 
			 	 int* pInt,
			 	 const int nDefault)
{
  aErr settingErr = aErrNone;
  char* pValue;

  aVALIDSETTINGFILE(pSettingFile);
  
  if (settingErr == aErrNone)
    settingErr = aSymbolTableInternal_Find(pSettingFile->pSettingTable,
  					   key, (void**)&pValue);

  switch (settingErr) {

  case aErrNone:
    aIntFromString(pInt, pValue);
    break;

  case aErrNotFound:
    *pInt = nDefault;
    settingErr = aErrNone;
    break;

  default:
    break;

  } /* switch */

  return settingErr;

} /* aSettingFileInternal_GetInt */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aSettingFileInternal_GetULong
 */

aErr aSettingFileInternal_GetULong(aSettingFile* pSettingFile,
			   	   const char* key, 
			   	   unsigned long* pULong,
			   	   const unsigned long nDefault)
{
  aErr settingErr = aErrNone;
  char* pValue;
  int val;

  aVALIDSETTINGFILE(pSettingFile);
  
  if (settingErr == aErrNone)
    settingErr = aSymbolTableInternal_Find(pSettingFile->pSettingTable,
  		    		 	   key, (void**)&pValue);

  switch (settingErr) {

  case aErrNone:
    aIntFromString(&val, pValue);
    *pULong = (unsigned long)val;
    break;

  case aErrNotFound:
    *pULong = nDefault;
    settingErr = aErrNone;
    break;

  default:
    break;

  } /* switch */

  return settingErr;

} /* aSettingFileInternal_GetULong */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aSettingFileInternal_GetFloat
 */

aErr aSettingFileInternal_GetFloat(aSettingFile* pSettingFile,
			   	   const char* key, 
			   	   float* pFloat,
			   	   const float fDefault)
{
  aErr settingErr = aErrNone;
  char* pValue;
  float value = 0.0f;

  aVALIDSETTINGFILE(pSettingFile);
  
  if (settingErr == aErrNone)
    settingErr = aSymbolTableInternal_Find(pSettingFile->pSettingTable,
  			    		   key, (void**)&pValue);

  if (settingErr == aErrNone)
    value = aString_ParseFloat(pValue);

  switch (settingErr) {

  case aErrNone:
    *pFloat = value;
    break;

  case aErrNotFound:
    *pFloat = fDefault;
    settingErr = aErrNone;
    break;

  default:
    break;

  } /* switch */

  return settingErr;

} /* aSettingFileInternal_GetFloat */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aSettingFileInternal_GetString
 */

aErr aSettingFileInternal_GetString(aSettingFile* pSettingFile,
			    	    const char* key, 
			    	    char** ppString,
			    	    const char* pDefault)
{
  aErr settingErr = aErrNone;
  void* pValue;

  aVALIDSETTINGFILE(pSettingFile);
  
  if (settingErr == aErrNone)
    settingErr = aSymbolTableInternal_Find(pSettingFile->pSettingTable,
  			    		   key, (void**)&pValue);

  switch (settingErr) {

  case aErrNone:
    *ppString = (char*)pValue;
    break;

  case aErrNotFound:
    *ppString = (char*)pDefault;
    settingErr = aErrNone;
    break;

  default:
    break;

  } /* switch */

  return settingErr;

} /* aSettingFileInternal_GetString */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aSettingFileInternal_GetInetAddr
 */

aErr aSettingFileInternal_GetInetAddr(aSettingFile* pSettingFile,
			    	      const char* key, 
			    	      unsigned long* pInetAddr,
			    	      const unsigned long pDefault)
{
  aErr settingErr = aErrNone;
  char* pValue;
  unsigned long address = 0;
  int num = 0;
  int byte = 0;
  int shift = 24;

  aVALIDSETTINGFILE(pSettingFile);
  
  if (settingErr == aErrNone)
    settingErr = aSymbolTableInternal_Find(pSettingFile->pSettingTable,
  			    		   key, (void**)&pValue);

  switch (settingErr) {

  case aErrNone:
    /* parse the float */
    while (*pValue && (settingErr == aErrNone)) {
      if (*pValue == '.') {
        if (shift < 8)
          settingErr = aErrParse;
        else {
          address += (unsigned long)(byte << shift);
          shift -= 8;;
          byte = 0;
        }
      } else if ((*pValue < '0') || (*pValue > '9')) {
        settingErr = aErrParse;
      } else {
        num = *pValue - '0';
        byte *= 10;
        byte += num;
        if (byte > 255)
          settingErr = aErrParse;
      }
      pValue++;
    }
    if (settingErr == aErrNone) {
      address += (unsigned long)byte;
      *pInetAddr = address;
    }
    break;

  case aErrNotFound:
    *pInetAddr = pDefault;
    settingErr = aErrNone;
    break;

  default:
    break;

  } /* switch */

  return settingErr;

} /* aSettingFileInternal_GetInetAddr */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aSettingFileInternal_Destroy
 */

aErr aSettingFileInternal_Destroy(aSettingFile* pSettingFile)
{
  aErr settingErr = aErrNone;

  aVALIDSETTINGFILE(pSettingFile);

  if (settingErr == aErrNone) {
    if (pSettingFile->pSettingTable != NULL) {
      aSymbolTableInternal_Destroy(pSettingFile->pSettingTable);
      pSettingFile->pSettingTable = NULL;
    }
    if (pSettingFile->pSettingPool != NULL) {
      aMemPoolInternal_Destroy(pSettingFile->pSettingPool);
      pSettingFile->pSettingPool = NULL;
    }
    pSettingFile->ioLib = NULL;
    pSettingFile->check = 0;
    aMemFree((aMemPtr)pSettingFile);
  }

  return settingErr;

} /* aSettingFileInternal_Destroy */

