/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                 */
/* file: aStemCore.c     	                                   */
/*                                                                 */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                 */
/* description: Implementation of core access routines for the 	   */
/*		BrainStem modules.	    			   */
/*                                                                 */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                 */
/* Copyright  1997-2002. Acroname Inc.                            */
/*                                                                 */
/* This software is the property of Acroname Inc.  Any             */
/* distribution, sale, transmission, or re-use of this code is     */
/* strictly forbidden except with permission from Acroname Inc.    */
/*                                                                 */
/* To the full extent allowed by law, Acroname Inc. also excludes  */
/* for itself and its suppliers any liability, wheither based in   */
/* contract or tort (including negligence), for direct,            */
/* incidental, consequential, indirect, special, or punitive       */
/* damages of any kind, or for loss of revenue or profits, loss of */
/* business, loss of information or data, or other financial loss  */
/* arising out of or in connection with this software, even if     */
/* Acroname Inc. has been advised of the possibility of such       */
/* damages.                                                        */
/*                                                                 */
/* Acroname Inc.                                                   */
/* www.acroname.com                                                */
/* 720-564-0373                                                    */
/*                                                                 */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "aCmd.tea"

#include "aUtil.h"
#include "aStemCore.h"

#define aLAUNCHVMTIMEOUTMS	100
#define aSERVOREPLYTIMEOUTMS	100
#define aA2DREADTIMEOUTMS	100



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * sCmdFilter
 */

aBool aCmdFilter(const unsigned char module,
		 const unsigned char dataLength,
		 const char* data,
		 void* ref)
{
  int temp = (int)ref;
  unsigned char cmd = (unsigned char)temp;

  if ((dataLength > 0) && (data[0] == cmd))
    return aTrue;

  return aFalse;

} /* sCmdFilter */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aStemCore_LaunchVM
 */

aErr aStemCore_LaunchVM(aStemLib stemLib,
			const unsigned char module,
			const unsigned char fileSlot,
			const char* pProgramData,
			const unsigned char nProgramDataLen,
        		const unsigned char flags,
			aTEAProcessID* pPID)
{
  aErr scErr = aErrNone;
  aPacketRef packet;
  unsigned char address;
  unsigned char length;
  char data[aSTEMMAXPACKETBYTES];
  unsigned char dataLen;
  char* p;
  unsigned char dataLeft = nProgramDataLen;
  int numPackets = 0;
  aTEAProcessID pid;
  aBool bDone = aFalse;

  data[0] = cmdVM_RUN;
  data[1] = 0;
  p = (char*)pProgramData;

  while ((scErr == aErrNone)
         && (bDone != aTrue)) {

    /* clear out the flag byte */
    data[1] = 0;
    length = 2;

    if (numPackets == 0) {
      /* set the first packet bit */
      data[1] |= bitVM_RUN_FIRST;
      
      /* set the file slot */
      data[2] = (char)fileSlot;
      length++;

      /* set the process ID if specified */
      if (flags & bitVM_RUN_PID) {
        /* set the file slot */
        data[1] |= bitVM_RUN_PID;
        data[3] = (char)*pPID;
        length++;
      } 

    /* packets beyond first need process id */
    } else {
      data[1] |= bitVM_RUN_PID;
      data[2] = (char)pid;
      length++;
    }

    /* add in any data */
    if (dataLeft > 0) {

      /* stuff as much data as possible */
      dataLen = (unsigned char)(aSTEMMAXPACKETBYTES - length);
      if (dataLeft < dataLen) {
        dataLen = dataLeft;
        dataLeft = 0;
      } else {
        dataLeft -= dataLen;
      }
      aMemCopy(&data[length], p, dataLen);
      p += dataLen;
      length += dataLen;
    }
 
    if (dataLeft == 0) {
      data[1] |= bitVM_RUN_LAST;
      bDone = aTrue;
    }

    /* now, build up the packet */
    aPacket_Create(stemLib, module, length, data, &packet, &scErr);
    
    /* send it */
    if (scErr == aErrNone)
      aStem_SendPacket(stemLib, packet, &scErr);
    
    /* wait for the reply */
    if (scErr == aErrNone)
      aStem_GetPacket(stemLib, aCmdFilter, (void*)cmdVM_RUN, 
      		      aLAUNCHVMTIMEOUTMS, &packet, &scErr);
    
    /* get the reply packet data */
    if (scErr == aErrNone)
      aPacket_GetData(stemLib, packet, &address,
      		      &length, data, &scErr);

    /* get the pid from the packet */
    if (scErr == aErrNone) {
      aAssert(length == 2);
      aAssert(data[0] == cmdVM_RUN);
      pid = (unsigned char)data[1];
    }
    
    /* clean up the packet */
    if (scErr == aErrNone)
      aPacket_Destroy(stemLib, packet, &scErr);
  }

  return scErr;
  
} /* aStemCore_LaunchVM */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aStem_GetA2D_Int
 */

aLIBRETURN aStem_GetA2D_Int(aStemLib stemLib,
			    const unsigned char module,
			    const unsigned char nAnalogIndex,
			    int *pAnalogValue,
			    aErr* pErr)
{
  aErr scErr = aErrNone;
  char data[aSTEMMAXPACKETBYTES];
  unsigned char address;
  unsigned char length;
  aPacketRef packet;

  if ((stemLib == NULL) || (pAnalogValue == NULL))
    scErr = aErrParam;

  /* send the outbound A/D request */  
  if (scErr == aErrNone) {
    data[0] = cmdA2D_RD;
    data[1] = (char)(nAnalogIndex | maskDEV_VAL_HOST);
    if (!aPacket_Create(stemLib, module, 2, data, &packet, &scErr))
      aStem_SendPacket(stemLib, packet, &scErr);
  }

  /* seek the reply */
  if (scErr == aErrNone) {
    if (!aStem_GetPacket(stemLib, aCmdFilter, (void*)cmdDEV_VAL,
      		         aA2DREADTIMEOUTMS, &packet, &scErr)
        && !aPacket_GetData(stemLib, packet, &address,
      		            &length, data, &scErr)) {
      if (length == 4) {
        unsigned short uval;
        uval = aUtil_RetrieveUShort(&data[2]);
        *pAnalogValue = (int)(uval >> 6);
      } else
        scErr = aErrNotFound;
      aPacket_Destroy(stemLib, packet, &scErr);
    }
  }

  return scErr;

} /* aStem_GetA2D_Int */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aStem_GetServoPosition
 */

aLIBRETURN aStem_GetServoPosition(aStemLib stemLib,
				  const unsigned char module,
				  const unsigned char nServoIndex,
				  unsigned char *pServoPosition,
				  aErr* pErr)
{
  aErr scErr = aErrNone;
  char data[aSTEMMAXPACKETBYTES];
  unsigned char address;
  unsigned char length;
  aPacketRef packet;

  aAssert(stemLib);
  aAssert(pServoPosition);

  data[0] = cmdSRV_ABS;
  data[1] = (char)nServoIndex;

  /* build and send the packet */
  aPacket_Create(stemLib, module, 2, data, &packet, &scErr);
  if (scErr == aErrNone)
    aStem_SendPacket(stemLib, packet, &scErr);

  /* look for the reply */
  if (scErr == aErrNone)
    aStem_GetPacket(stemLib, aCmdFilter, (void*)cmdSRV_ABS, 
      		    aSERVOREPLYTIMEOUTMS, &packet, &scErr);

  /* get the packet data */
  if (scErr == aErrNone)
    aPacket_GetData(stemLib, packet, &address, &length, data, &scErr);

  if (scErr == aErrNone) {
    aAssert(data[0] == cmdSRV_ABS);
    aAssert(data[1] == nServoIndex);
    *pServoPosition = (unsigned char)data[2];
    aPacket_Destroy(stemLib, packet, &scErr);
  }

  if (pErr != NULL)
    *pErr = scErr;

  return (aLIBRETURN)(scErr != aErrNone);

} /* aStem_GetServoPosition */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aStem_GetServoConfig
 */

aLIBRETURN aStem_GetServoConfig(aStemLib stemLib,
				const unsigned char module,
				const unsigned char nServoIndex,
				unsigned char *pConfigByte,
				aErr* pErr)
{
  aErr scErr = aErrNone;
  char data[aSTEMMAXPACKETBYTES];
  unsigned char address;
  unsigned char length;
  aPacketRef packet;

  aAssert(stemLib);
  aAssert(pConfigByte);  

  data[0] = cmdSRV_CFG;
  data[1] = (char)nServoIndex;

  /* build and send the packet */
  aPacket_Create(stemLib, module, 2, data, &packet, &scErr);
  if (scErr == aErrNone)
    aStem_SendPacket(stemLib, packet, &scErr);

  /* look for the reply */
  if (scErr == aErrNone)
    aStem_GetPacket(stemLib, aCmdFilter, (void*)cmdSRV_CFG, 
      		    aSERVOREPLYTIMEOUTMS, &packet, &scErr);

  /* get the packet data */
  if (scErr == aErrNone)
    aPacket_GetData(stemLib, packet, &address, &length, data, &scErr);

  if (scErr == aErrNone) {
    aAssert(data[0] == cmdSRV_CFG);
    aAssert(data[1] == nServoIndex);
    *pConfigByte = (unsigned char)data[2];
    aPacket_Destroy(stemLib, packet, &scErr);
  }

  if (pErr != NULL)
    *pErr = scErr;

  return (aLIBRETURN)(scErr != aErrNone);

} /* aStem_GetServoConfig */



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aStem_GetServoLimits
 */

aLIBRETURN aStem_GetServoLimits(aStemLib stemLib,
				const unsigned char module,
				const unsigned char nServoIndex,
				unsigned char *pLimitBytes,
				aErr* pErr)
{
  aErr scErr = aErrNone;
  char data[aSTEMMAXPACKETBYTES];
  unsigned char address;
  unsigned char length;
  aPacketRef packet;

  aAssert(stemLib);
  aAssert(pLimitBytes);  

  data[0] = cmdSRV_LMT;
  data[1] = (char)nServoIndex;

  /* build and send the packet */
  aPacket_Create(stemLib, module, 2, data, &packet, &scErr);
  if (scErr == aErrNone)
    aStem_SendPacket(stemLib, packet, &scErr);

  /* look for the reply */
  if (scErr == aErrNone)
    aStem_GetPacket(stemLib, aCmdFilter, (void*)cmdSRV_LMT, 
      		    aSERVOREPLYTIMEOUTMS, &packet, &scErr);

  /* get the packet data */
  if (scErr == aErrNone)
    aPacket_GetData(stemLib, packet, &address, &length, data, &scErr);

  if (scErr == aErrNone) {
    aAssert(data[0] == cmdSRV_LMT);
    aAssert(data[1] == nServoIndex);
    aMemCopy(pLimitBytes, &data[2], 2);
    aPacket_Destroy(stemLib, packet, &scErr);
  }

  if (pErr != NULL)
    *pErr = scErr;

  return (aLIBRETURN)(scErr != aErrNone);

} /* aStem_GetServoLimits */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aStem_SetServoPositionAbs
 */

aLIBRETURN aStem_SetServoPositionAbs(aStemLib stemLib,
				     const unsigned char module,
				     const unsigned char nServoIndex,
				     const unsigned char nPosition,
				     aErr* pErr)
{
  aErr scErr = aErrNone;
  char data[aSTEMMAXPACKETBYTES];
  aPacketRef packet;

  aAssert(stemLib);

  data[0] = cmdSRV_ABS;
  data[1] = (char)nServoIndex;
  data[2] = (char)nPosition;

  aPacket_Create(stemLib, module, 3, data, &packet, &scErr);

  if (scErr == aErrNone)
    aStem_SendPacket(stemLib, packet, &scErr);

  if (pErr != NULL)
    *pErr = scErr;

  return (aLIBRETURN)(scErr != aErrNone);

} /* aStem_SetServoPositionAbs */



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aStem_SetServoPositionRel
 */

aLIBRETURN aStem_SetServoPositionRel(aStemLib stemLib,
				     const unsigned char module,
				     const unsigned char nServoIndex,
				     const unsigned char nChange,
				     aErr* pErr)
{
  aErr scErr = aErrNone;
  char data[aSTEMMAXPACKETBYTES];
  aPacketRef packet;

  aAssert(stemLib);

  data[0] = cmdSRV_ABS;
  data[1] = (char)nServoIndex;
  data[2] = (char)nChange;

  aPacket_Create(stemLib, module, 3, data, &packet, &scErr);

  if (scErr == aErrNone)
    aStem_SendPacket(stemLib, packet, &scErr);

  if (pErr != NULL)
    *pErr = scErr;

  return (aLIBRETURN)(scErr != aErrNone);

} /* aStem_SetServoPositionRel */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aStem_SetServoConfig
 */

aLIBRETURN aStem_SetServoConfig(aStemLib stemLib,
				const unsigned char module,
				const unsigned char nServoIndex,
				const unsigned char configByte,
				aErr* pErr)
{
  aErr scErr = aErrNone;
  char data[aSTEMMAXPACKETBYTES];
  aPacketRef packet;

  aAssert(stemLib);

  data[0] = cmdSRV_CFG;
  data[1] = (char)nServoIndex;
  data[2] = (char)configByte;

  aPacket_Create(stemLib, module, 3, data, &packet, &scErr);

  if (scErr == aErrNone)
    aStem_SendPacket(stemLib, packet, &scErr);

  if (pErr != NULL)
    *pErr = scErr;

  return (aLIBRETURN)(scErr != aErrNone);

} /* aStem_SetServoConfig */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aStem_SetServoLimits
 */

aLIBRETURN aStem_SetServoLimits(aStemLib stemLib,
				const unsigned char module,
				const unsigned char nServoIndex,
				const unsigned char *pLimitBytes,
				aErr* pErr)
{
  aErr scErr = aErrNone;
  char data[aSTEMMAXPACKETBYTES];
  aPacketRef packet;

  aAssert(stemLib);
  aAssert(pLimitBytes);  

  data[0] = cmdSRV_LMT;
  data[1] = (char)nServoIndex;
  aMemCopy(&data[2], pLimitBytes, 2);

  /* build and send the packet */
  aPacket_Create(stemLib, module, 4, data, &packet, &scErr);
  if (scErr == aErrNone)
    aStem_SendPacket(stemLib, packet, &scErr);
  
  if (pErr != NULL)
    *pErr = scErr;

  return (aLIBRETURN)(scErr != aErrNone);

} /* aStem_SetServoLimits */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aServo_CommitSettings
 */

aLIBRETURN aServo_CommitSettings(aStemLib stemLib,
				 const unsigned char module,
				 aErr* pErr)
{
  aErr scErr = aErrNone;
  char data[aSTEMMAXPACKETBYTES];
  aPacketRef packet;

  aAssert(stemLib);

  data[0] = cmdSRV_SAV;

  /* build and send the packet */
  aPacket_Create(stemLib, module, 1, data, &packet, &scErr);
  if (scErr == aErrNone)
    aStem_SendPacket(stemLib, packet, &scErr);
  
  if (pErr != NULL)
    *pErr = scErr;

  return (aLIBRETURN)(scErr != aErrNone);

} /* aServo_CommitSettings */
