/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                 */
/* file: aStemInternal.c                                           */
/*                                                                 */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                 */
/* description: Definition of a platform-independent BrainStem	   */
/*		communication layer.    			   */
/*                                                                 */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                 */
/* Copyright  1997-2000. Acroname Inc.                            */
/*                                                                 */
/* This software is the property of Acroname Inc.  Any             */
/* distribution, sale, transmission, or re-use of this code is     */
/* strictly forbidden except with permission from Acroname Inc.    */
/*                                                                 */
/* To the full extent allowed by law, Acroname Inc. also excludes  */
/* for itself and its suppliers any liability, wheither based in   */
/* contract or tort (including negligence), for direct,            */
/* incidental, consequential, indirect, special, or punitive       */
/* damages of any kind, or for loss of revenue or profits, loss of */
/* business, loss of information or data, or other financial loss  */
/* arising out of or in connection with this software, even if     */
/* Acroname Inc. has been advised of the possibility of such       */
/* damages.                                                        */
/*                                                                 */
/* Acroname Inc.                                                   */
/* www.acroname.com                                                */
/* 720-564-0373                                                    */
/*                                                                 */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifdef aPALM
#include <SysEvtMgr.h> /* for resetting auto off on heartbeat */
#endif /* aPALM */

#include "aStem.h"
#include "aStemText.h"
#include "aStemInternal.h"


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * local prototypes
 */

aErr aStemInternal_LogDebugPacket(aStem* pStem,
				  aPacket* pPacket);


aErr aStemInternal_HandleHB(aStem* pStem,
			    aPacket* pPacket,
			    aPacketFilter filterProc,
			    void* filterRef,
			    aPacketList* pPacketList);

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aStemInternal_LogDebugPacket
 */

aErr aStemInternal_LogDebugPacket(aStem* pStem,
				  aPacket* pPacket)
{
  aErr stemErr = aErrNone;

  /* add the packet to the debug stream if present */
  if (pStem->debugStream != NULL) {
    aErr dbgErr;
    char debug[aSTEMMAXDEBUGLINE];
    dbgErr = aPacketInternal_Format(pPacket, debug, 
    				    aSTEMMAXDEBUGLINE);
    if (dbgErr == aErrNone)
      aStream_WriteLine(aStreamLibRef(pStem->debugStream),
                	pStem->debugStream, debug, NULL);
  }

  return stemErr;

} /* aStemInternal_LogDebugPacket */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aStemInternal_Create
 */

aErr aStemInternal_Create(aStem** ppStem)
{
  aStem* pStem = NULL;
  aErr stemErr = aErrNone;
  
  if (ppStem == NULL)
    stemErr = aErrParam;

  /* build the stem object */
  if (stemErr == aErrNone) {
    pStem = (aStem*)aMemAlloc(sizeof(aStem));
    if (pStem == NULL)
      stemErr = aErrMemory;
    else {
      aBZero(pStem, sizeof(aStem));
      pStem->check = aSTEMCHECK;
    }
  }

  /* open up the io library */
  if (stemErr == aErrNone)
    aIO_GetLibRef(&pStem->ioRef, &stemErr);
  
  /* now, set up the stem */

  if (stemErr == aErrNone) {
    aPacket_Init(&pStem->curLinkPacket, aPacketFromStem);
    aPacket_Init(&pStem->curRelayPacket, 0);
    if (stemErr == aErrNone)
      aMemPool_Create(pStem->ioRef, sizeof(aPacket), 
        	      aPACKETPOOLBLOCKSIZE, &pStem->packetPoolRef,
        	      &stemErr);
    if (stemErr == aErrNone)
      stemErr = aPacketList_Create(pStem->packetPoolRef,
        		   	   &pStem->pLinkPacketList);
    if (stemErr == aErrNone)
      stemErr = aPacketList_Create(pStem->packetPoolRef,
        		   	   &pStem->pRelayPacketList);
    if (stemErr == aErrNone)
      *ppStem = pStem;
    else
      aStemInternal_Destroy(pStem);
  }

  /* try to find and setup a debug file */
  if (stemErr == aErrNone) {
    aErr ioErr;
    aFileRef testFile;
    
    aAssert(pStem->ioRef);

    /* check for debug file */
    aFile_Open(pStem->ioRef, aSTEM_DEBUGFILENAME, 
  	       aFileModeReadOnly, aFileAreaBinary,
  	       &testFile, &ioErr);

    /* just close it if you found it */
    if (ioErr == aErrNone)
      aFile_Close(pStem->ioRef, testFile, &ioErr);

    /* then open it for writing if it is there */
    if (ioErr == aErrNone)
      aStream_CreateFileOutput(pStem->ioRef, aSTEM_DEBUGFILENAME,
        		       aFileAreaBinary, &pStem->debugStream, 
      			       &ioErr);

    /* write a preamble */
    if (ioErr == aErrNone)
      aStream_WriteLine(pStem->ioRef, pStem->debugStream,
      			aSTEM_DBG_LIBOPEN, &ioErr);
  }

  return(stemErr);

} /* aStemInternal_Create */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aStemInternal_SetStream
 */

aErr aStemInternal_SetStream(aStem* pStem, 
			     aStreamRef streamRef,
			     aStemStreamType streamType)
{
  aErr stemErr = aErrNone;
  aStreamRef* pStream = NULL;

  aVALIDSTEM(pStem);

  /* get the stream in question */
  switch (streamType) {
  case kStemModuleStream:
    pStream = &pStem->linkStream;
    if (pStem->debugStream) {
      aStream_WriteLine(aStreamLibRef(pStem->debugStream),
      			pStem->debugStream,
      			aSTEM_DBG_MODULESTREAM, NULL);
    }
    break;
  case kStemRelayStream:
    pStream = &pStem->relayStream;
    if (pStem->debugStream) {
      aStream_WriteLine(aStreamLibRef(pStem->debugStream),
      			pStem->debugStream,
      			aSTEM_DBG_RELAYSTREAM, NULL);
    }
    break;
  } /* switch */

  if (pStream == NULL)
    stemErr = aErrParam;

  /* throw away any old stream being used */
  if ((stemErr == aErrNone) 
      && (pStream != NULL)
      && (*pStream != NULL)) {
    aStreamRef stream = *pStream;
    *pStream = NULL;
    if (pStem->debugStream) {
      aStream_WriteLine(aStreamLibRef(pStem->debugStream),
      			pStem->debugStream,
      			aSTEM_DBG_STREAMCLEARED, NULL);
    }
    aStream_Destroy(aStreamLibRef(stream),
      		    stream, 
      		    &stemErr);
    if ((stemErr == aErrNone) 
        && (pStem->debugStream)) {
      aStream_WriteLine(aStreamLibRef(pStem->debugStream),
      			pStem->debugStream,
      			aSTEM_DBG_STREAMDISPOSED, NULL);
    }
  }

  /* set the new stream */
  if ((stemErr == aErrNone) 
      && (streamRef != NULL)) {
    *pStream = streamRef;
    if (pStem->debugStream) {
      aStream_WriteLine(aStreamLibRef(pStem->debugStream),
      			pStem->debugStream,
      			aSTEM_DBG_STREAMSET, NULL);
    }
  }

  return stemErr;

} /* aStemInternal_SetStream */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aStemInternal_DebugLine
 */

aErr aStemInternal_DebugLine(aStem* pStem,
			     const char* line)
{
  aErr stemErr = aErrNone;
  
  aVALIDSTEM(pStem);

  if (line == NULL)
    stemErr = aErrParam;

  if ((stemErr == aErrNone) 
      && (pStem->debugStream != NULL))
    aStream_WriteLine(aStreamLibRef(pStem->debugStream),
    		      pStem->debugStream, line, &stemErr);

  return stemErr;

} /* aStemInternal_DebugLine */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aStemInternal_DrainStream
 *
 * This function drains out the input stream of any inbound bytes
 * and then returns.  There is no error if there are no packets 
 * available.
 */

aErr aStemInternal_DrainStream(aStem* pStem,
			       aStreamRef stream,
			       aPacket* pPacketStore,
			       aPacketFilter filterProc,
			       void* filterRef,
			       aBool bHandleHB,
			       unsigned char initflags,
			       aPacketList* pPacketList)
{
  aErr stemErr = aErrNone;
  aErr readErr = aErrNone;
  char byte;

  aAssert(pStem);

  if ((stemErr == aErrNone) 
      && (stream != NULL)) {
    do {
      if (!aStream_Read(aStreamLibRef(stream), 
      		        stream, &byte, 1, &readErr)) {

        stemErr = aPacket_AddByte(pPacketStore, 
        			  (unsigned char)byte);
        if ((stemErr == aErrNone)
            && (aPacket_Complete(pPacketStore) == aTrue)) {

	  /* record reciept of the packet */
          aStemInternal_LogDebugPacket(pStem, pPacketStore);

	  /* see if it is a heartbeat packet we should handle */
          if (stemErr == aErrNone) {
            aErr hbErr;
            if (bHandleHB == aTrue)
              hbErr = aStemInternal_HandleHB(pStem, pPacketStore,
            				     filterProc, filterRef,
      				             pStem->pLinkPacketList);
      	    else
      	      hbErr = aErrNotFound;

            /* all non-hb packets get added to the list */
      	    if (hbErr == aErrNotFound) {
              stemErr = aPacketList_AddCopy(pStem, 
              				    pPacketList, 
              				    pPacketStore);

            /* show an error if we didn't handle the hb */
            } else if (hbErr != aErrNone)
              stemErr = hbErr;
          }

          /* clean up and start with a fresh packet */
          aPacket_Init(pPacketStore, initflags);
        }
      }
    } while ((stemErr == aErrNone) && (readErr == aErrNone));
  }

  return(stemErr);

} /* aStemInternal_DrainStream */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aStemInternal_GetPacket
 *
 * Strategy:
 *
 * We need to follow a prescribed path to support all the cases.
 * First, we need to flush the stream of all bytes and accumulate 
 * these bytes into packets in the packet list.
 *
 * Now there are two options, filtered and not filtered.
 *
 *  Filtered: In this case, we walk the list applying the 
 *            filter to look for a match.  If the filter
 *            returns true, we stop looking and return 
 *	      the packet with one exception.  If the 
 *	      packet is a heartbeat packet, we service
 *	      the packet, don't give it back, and return 
 *	      an aErrNotFound message.
 *
 * Unfiltered: Here, the entire list is walked to service
 *	      (and remove) the heartbeat packets and we 
 *	      return the packet at the head of the list 
 *	      if it is present.
 *
 * We want to handle the following cases:
 *
 * 1. no filter, several packets arrive, we cleans the
 *    list of the heartbeats (to keep things running 
 *    smoothly) and return the first packet that was 
 *    available.
 * 2. no filter, only one packet that is a heartbeat
 *    we service the heartbeat and return aErrNotFound.
 * 3. filter, we apply the filter to find the first
 *    match but we also handle all heartbeat packets
 * 4. filter, none match, we still clean out the 
 *    hearbeats and return aErrNotFound
 *
 * Basically, we always process any available heartbeats
 * since we have the chance and the heartbeats never 
 * get returned.  If a filter is present, it sees 
 * every packet, including heartbeats.
 */

aErr aStemInternal_GetPacket(aStem* pStem,
			     aPacketFilter filterProc,
			     void* filterRef,
			     unsigned long nMSTimeout,
			     aPacket** ppPacket)
{
  aErr stemErr = aErrNone;
  unsigned long nStart, nCurrent;

  aAssert(pStem);
  aAssert(pStem->pLinkPacketList);
  aAssert(ppPacket);

  aVALIDSTEM(pStem);

  /* be sure we have a valid packet return value */
  if ((stemErr == aErrNone) && (ppPacket == NULL))
    stemErr = aErrParam;

  /* just say no if no stream has been set */
  if ((stemErr == aErrNone) && (pStem->linkStream == NULL))
    stemErr = aErrNotReady;

  /* establish the start time for timeouts */
  if (stemErr == aErrNone)
    aIO_GetMSTicks(pStem->ioRef, &nStart, &stemErr);

  /* if not relaying, try to filter and retrieve the packet */
  if (pStem->relayStream == NULL) {

    /* if a filter proc was set, continue to request packets
     * until the filter finds what it wants or a timeout 
     * occurs
     */
    find_packet:

    /* suck in any available packets into the packet list, regardless
     * of what type they are */
    if (stemErr == aErrNone) {
      stemErr = aStemInternal_DrainStream(pStem,
      					  pStem->linkStream,
      					  &pStem->curLinkPacket,
      					  filterProc, filterRef,
      					  aTrue,
      					  aPacketFromStem,
      					  pStem->pLinkPacketList);
    }

    /* see if one is available in the list and return it */
    if (stemErr == aErrNone)
      stemErr = aPacketList_GetFirst(pStem->pLinkPacketList, 
    				     filterProc, filterRef,
    				     ppPacket);

    /* if the timeout expired, stop trying */
    if (stemErr == aErrNotFound) {
      aIO_GetMSTicks(pStem->ioRef, &nCurrent, &stemErr);
      if (nCurrent - nStart >= nMSTimeout) {
        stemErr = aErrTimeout;
      } else {
        stemErr = aErrNone;
        goto find_packet;
      }
    }

  /* if a relay is set, just get any packets and relay */
  } else {
    aPacket* pPacket;

    /* drain both streams */
    if (stemErr == aErrNone) {
      stemErr = aStemInternal_DrainStream(pStem, 
      					  pStem->linkStream,
      					  &pStem->curLinkPacket,
      					  filterProc, filterRef,
      					  aFalse,
      					  aPacketFromStem,
      					  pStem->pLinkPacketList);
    } 

    if (stemErr == aErrNone) {
      stemErr = aStemInternal_DrainStream(pStem, 
      					  pStem->relayStream,
      					  &pStem->curRelayPacket,
      					  filterProc, filterRef,
      					  aFalse,
      					  0,
      					  pStem->pRelayPacketList);

    }

    /* check for module -> relay packets and send along */
    if (stemErr == aErrNone) {
      stemErr = aPacketList_GetFirst(pStem->pLinkPacketList, 
    				     filterProc, filterRef,
    				     &pPacket);   
      if (stemErr == aErrNone) {
        stemErr = aPacketInternal_WriteToStream(pPacket, 
    					        pStem->relayStream);
        aPacketInternal_Destroy(pStem, pPacket);
      } else if (stemErr == aErrNotFound) {
        stemErr = aErrNone;
      }
    }

    /* check for relay -> module packets */
    if (stemErr == aErrNone) {
      stemErr = aPacketList_GetFirst(pStem->pRelayPacketList, 
    				     filterProc, filterRef, 
    				     &pPacket);
      if (stemErr == aErrNone) {

        /* we let the filterProc see the heartbeats on the way through */
        if (aPACKET_IS_HB(pPacket))
          filterProc(pPacket->address,
          	     pPacket->curSize,
          	     pPacket->data,
          	     filterRef);

        stemErr = aPacketInternal_WriteToStream(pPacket, 
    					        pStem->linkStream);
        aPacketInternal_Destroy(pStem, pPacket);
      } else if (stemErr == aErrNotFound) {
        stemErr = aErrNone;
      }
    }

    /* check to see if the relay was dropped and clear the relay */
    if (stemErr != aErrIO)
      /* when relaying, we never get packets */
      stemErr = aErrNotFound;
  }

  return stemErr;

} /* aStemInternal_GetPacket */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aStemInternal_SendPacket
 */

aErr aStemInternal_SendPacket(aStem* pStem,
			      aPacket* pPacket)
{
  aErr stemErr = aErrNone;

  aAssert(pStem);
  aAssert(pPacket);
  aAssert(pPacket->dataSize <= aSTEMMAXPACKETBYTES);

  aVALIDSTEM(pStem);
  aVALIDPACKET(pPacket);

  /* make sure we have a data size */
  if ((stemErr == aErrNone) && 
      (pPacket->dataSize > aSTEMMAXPACKETBYTES))
    stemErr = aErrRange;

  /* make sure we have a stream */
  if ((stemErr == aErrNone) && (pStem->linkStream == NULL))
    stemErr = aErrIO;
  
  /* send the packet down the stream */
  if (stemErr == aErrNone)
    stemErr = aPacketInternal_WriteToStream(pPacket, 
    					    pStem->linkStream);

  /* output the debug packet if debug stream present */
  aStemInternal_LogDebugPacket(pStem, pPacket);

  /* always clean up the packet */
  aMemPool_Free(pStem->ioRef, 
  		pStem->packetPoolRef, 
    		pPacket, 
    		NULL);

  return stemErr;

} /* aStemInternal_SendPacket */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aStemInternal_Destroy
 */

aErr aStemInternal_Destroy(aStem* pStem)
{
  aErr stemErr = aErrNone;
  aErr ioErr;

  aVALIDSTEM(pStem);

  /* close down the relay stream */
  if (stemErr == aErrNone)
    ioErr = aStemInternal_SetStream(pStem, NULL, kStemRelayStream);

  /* close down the module stream */
  if (stemErr == aErrNone)
    ioErr = aStemInternal_SetStream(pStem, NULL, kStemModuleStream);

  /* close up the debug file if it is open */
  if ((stemErr == aErrNone)
      && (pStem->debugStream != NULL)) {
    aStream_WriteLine(pStem->ioRef, 
      		      pStem->debugStream,
      		      aSTEM_DBG_LIBCLOSE,
      		      &ioErr);
    aStream_Destroy(aStreamLibRef(pStem->debugStream), 
		    pStem->debugStream, &ioErr);
  }

  /* clean up the packet lists */
  if ((stemErr == aErrNone) && (pStem->pRelayPacketList))
    stemErr = aPacketList_Destroy(pStem, pStem->pRelayPacketList);
  if ((stemErr == aErrNone) && (pStem->pLinkPacketList))
    stemErr = aPacketList_Destroy(pStem, pStem->pLinkPacketList);

  if ((stemErr == aErrNone) && (pStem->packetPoolRef))
    aMemPool_Destroy(pStem->ioRef, pStem->packetPoolRef, &stemErr);

  /* clean up any reference we have to the io library */
  if ((stemErr == aErrNone)
      && (pStem->ioRef != NULL))
    aIO_ReleaseLibRef(pStem->ioRef, &ioErr);

  if (stemErr == aErrNone)
    aMemFree((aMemPtr)pStem);

  return(stemErr);

} /* aStemInternal_Destroy */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aStemInternal_HandleHB
 *
 * checks the packet and responds to it if it is a heartbeat.
 * aErrNone means it was a heartbeat, aErrNotFound means it wasn't.
 * other errors where appropriate
 */

aErr aStemInternal_HandleHB(aStem* pStem,
			    aPacket* pPacket,
			    aPacketFilter filterProc,
			    void* filterRef,
			    aPacketList* pPacketList)
{
  aErr stemErr = aErrNone;

  /* automatically respond to heatbeat packets here */
  if (aPACKET_IS_HB(pPacket)) {
    aPacket* pCopy;

    /* let the filter see the packet (we don't care what the 
     * return value for filtered heartbeat packets is */
    if (filterProc != NULL)
      filterProc(pPacket->address, pPacket->dataSize, 
      		 pPacket->data, filterRef);

    if (pStem->relayStream == NULL) {

      /* build a copy that will be disposed when sent */    
      aMemPool_Alloc(pStem->ioRef, pPacketList->packetPoolRef, 
    		     (void**)&pCopy, &stemErr);

      if (stemErr == aErrNone) {
        *pCopy = *pPacket;

        /* on PalmOS, we reset the sleep timer so the Palm won't 
         * sleep if there is an active heartbeat */
#ifdef aPALM
        EvtResetAutoOffTimer();
#endif /* aPALM */

        /* just send back the same packet modified to be a reply */
        pCopy->data[1] += 2;
        pCopy->status &= (unsigned char)~aPacketFromStem;
        stemErr = aStemInternal_SendPacket(pStem, pCopy);
      }
    } else {
      stemErr = aErrNotFound;
    }

  /* otherwise, just return that it was not a HB packet */
  } else {
    stemErr = aErrNotFound;
  }
  
  return stemErr;

} /* aStemInternal_HandleHB */
