/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                 */
/* file: aStemTEAFile.c                                 	   */
/*                                                                 */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                 */
/* description: cross-platform TEA file stream                     */
/*              implementation.                                    */
/*                                                                 */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                 */
/* Copyright  1997-2002. Acroname Inc.                            */
/*                                                                 */
/* This software is the property of Acroname Inc.  Any             */
/* distribution, sale, transmission, or re-use of this code is     */
/* strictly forbidden except with permission from Acroname Inc.    */
/*                                                                 */
/* To the full extent allowed by law, Acroname Inc. also excludes  */
/* for itself and its suppliers any liability, wheither based in   */
/* contract or tort (including negligence), for direct,            */
/* incidental, consequential, indirect, special, or punitive       */
/* damages of any kind, or for loss of revenue or profits, loss of */
/* business, loss of information or data, or other financial loss  */
/* arising out of or in connection with this software, even if     */
/* Acroname Inc. has been advised of the possibility of such       */
/* damages.                                                        */
/*                                                                 */
/* Acroname Inc.                                                   */
/* www.acroname.com                                                */
/* 720-564-0373                                                    */
/*                                                                 */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


#include "aCmd.tea"

#include "aStemInternal.h"

#define aTEAFILETIMEOUTMS		2000


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * structure used by the stream interface in the create input/output
 * routines for TEA program files.
 */

typedef struct aTEAFileState {

  /* we communicate through the stem stream */
  aStem*		pStem;
  
  /* holds the bytes being assembled for sending or bytes received
   * in the last packet that have not been used yet */
  char			packetBuffer[aSTEMMAXPACKETBYTES];

  /* tells how many bytes are valid in the buffer */
  int			nBytesInBuffer;
  
  /* checksum used to ensure the file is transfered safely */
  unsigned char		checksum;
  
  /* used to track page boundary stuff */
  int			nBytesSent;
  
  /* set to true when the cmdTEAEOF command is received */
  aBool			bDone;

  /* address of the module being accessed in subsequent
   * cmdTEADATA or cmdTEAREAD calls */
  unsigned char		address;
  
  /* file number being written */
  int 			nFileNumber;

} aTEAFileState;




/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * static local routines (callbacks)
 */

static aErr sStemTEAFile_Get(char* pData,
                             void* ref);
static aErr sStemTEAFile_Put(char* pData,
                             void* ref);
static aErr sStemTEAFile_DeleteOutput(void* ref);
static aErr sStemTEAFile_DeleteInput(void* ref);

static aBool sInitFilter(const unsigned char address,
			 const unsigned char length,
			 const char* data);
static aBool sCheckSumFilter(const unsigned char address,
			     const unsigned char length,
			     const char* data);
static aBool sDataFilter(const unsigned char address,
			 const unsigned char length,
			 const char* data);

static aErr sStemTEAFile_Init(aTEAFileState* pFileState);


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aStemInternal_CreateTEAFileInput
 * 
 * This routine creates an input stream to allow reading a TEA
 * program from a BrainStem Module.
 *
 * When this routine is called, it builds a stream interface using
 * aStream_Create. It also allocates (using aMemAlloc) a 
 * aTEAFileState structure that is defined at the top of this file.  
 * This structure is pointed to by the getRef and deleteRef 
 * of the stream interface and used by the get and delete 
 * callbacks.
 *
 * Opening the stream sends a cmdTEAINIT command packet to the 
 * addressed module to initiate the reading. The first packet is 
 * also requested using the cmdTEAREAD command.  When the first 
 * packet is obtained (using the aStemInternal_GetPacket routine), 
 * the checksum is updated and the packet bytes are stored in the 
 * state structure data buffer.
 *
 * Each get callback takes the next byte from the buffer and 
 * returns it.  When no more bytes are available in the buffer,
 * the next read packet is requested using the cmdTEAREAD command
 * within the get callback proc.
 *
 * When the last byte in the last packet is sent by the addressed
 * BrainStem module, the cmdTEAEOF packet is also sent by the 
 * module and the bDone boolean is set in the aTEAFileState 
 * structure.
 *
 * Subsequent reads will call get and it will immediately return
 * a aErrEOF error when the bDone boolean has been set.
 *
 * If the stream is destroyed before the bDone command is set, the
 * cmdTEAEOF command is sent to the module to reset the file state
 * and terminate the read operation.  The delete proc then frees 
 * up the stream state structure (type aTEAFileState) and returns
 * gracefully.
 */

aErr aStemInternal_CreateTEAFileInput(aStem* pStem, 
				      const unsigned char address,
				      const int nFileNumber,
				      aStreamRef* pStreamRef)
{
  aErr stemErr = aErrNone;
  aStreamRef streamRef;
  aTEAFileState* pPrivate;

  aVALIDSTEM(pStem);

  /* make sure we have a stream ref pointer and an even address */  
  if (pStreamRef == NULL)
    stemErr = aErrParam;
  else if (address & 0x01)
    stemErr = aErrParam;

  /* now build the private structure */  
  if (stemErr == aErrNone) {
    pPrivate = (aTEAFileState*)aMemAlloc(sizeof(aTEAFileState));

    if (pPrivate == NULL)
      stemErr = aErrMemory;
    else
      aBZero(pPrivate, sizeof(aTEAFileState));
  }
  
  /* fill in the structures */
  if (stemErr == aErrNone) {      
    pPrivate->pStem = pStem;
    pPrivate->address = address;
    pPrivate->nFileNumber = nFileNumber;
  } 

  /* build the generic stream */
  if (stemErr == aErrNone) {

    if (pStreamRef == NULL)
      return aErrParam;

    /* initialize the return value */
    *pStreamRef = NULL;

    aStream_Create(pStem->ioRef, 
    		   sStemTEAFile_Get,
    		   NULL,
    		   sStemTEAFile_DeleteInput,
    		   pPrivate,
    		   &streamRef, 
    		   &stemErr);
  }

  if (stemErr == aErrNone)
    *pStreamRef = streamRef;

  /* send and verify the init packet */
  if (stemErr == aErrNone)
    stemErr = sStemTEAFile_Init(pPrivate);

  return stemErr;

} /* aStemInternal_CreateTEAFileInput */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aStemInternal_CreateTEAFileOutput
 * 
 * This routine creates an output stream to allow writing a TEA
 * program to a BrainStem Module.
 *
 * The routine is very similar to the input analog but has the 
 * different sequence:
 *  - cmdTEAINIT initiates the write on the correct file of the 
 *               correct module
 *  - cmdTEADATA sends the packets to the stem
 *  - cmdTEAEOF signals the end of the transmission
 *
 * This routines builds a stream interface structure and a 
 * aTEAFileState structure but it only sets the put and delete
 * callback procs in the stream interface structure.
 * 
 * This routine may need to implement a delay to wait for the 
 * EEPROM write on the module. Chances are, the baud rate and 
 * packet size will allow enough time between cmdTEADATA packets
 * for the EEPROM write to occur.
 */

aErr aStemInternal_CreateTEAFileOutput(aStem* pStem, 
				       const unsigned char address,
				       const int nFileNumber,
				       aStreamRef* pStreamRef)
{
  aErr stemErr = aErrNone;
  aStreamRef streamRef = NULL;
  aTEAFileState* pPrivate = NULL;


  aVALIDSTEM(pStem);
  
  /* make sure we have a stream ref pointer, an even address, and 
   * an address in the correct range 
   */
  if (stemErr == aErrNone) {
    if (pStreamRef == NULL)
      stemErr = aErrParam;
    else if ((address & 0x01) || (address == 0x00))
      stemErr = aErrParam;
  }

  /* now build the private structure */  
  if (stemErr == aErrNone) {
    pPrivate = (aTEAFileState*)aMemAlloc(sizeof(aTEAFileState));

    if (pPrivate == NULL)
      stemErr = aErrMemory;
  }
  
  /* fill in the structures */
  if (stemErr == aErrNone) {
    aBZero(pPrivate, sizeof(aTEAFileState));
    pPrivate->pStem = pStem;
    pPrivate->address = address;
    pPrivate->nFileNumber = nFileNumber;
  }

  /* send and verify the init packet */
  if (stemErr == aErrNone)
    stemErr = sStemTEAFile_Init(pPrivate);

  /* if there were errors, clean up the memory allocated above */
  if ((stemErr != aErrNone) && (pPrivate != NULL))
    aMemFree((aMemPtr)pPrivate);

  /* build the generic stream */
  if (stemErr == aErrNone) {
    aAssert(pPrivate);

    /* initialize */
    *pStreamRef = NULL;
    if (aStream_Create(pStem->ioRef, 
    		       NULL,
    		       sStemTEAFile_Put,
    		       sStemTEAFile_DeleteOutput,
    		       pPrivate,
    		       &streamRef, 
    		       &stemErr))
      return stemErr;
  }
 
  /* set the output variable */
  if (stemErr == aErrNone)
    *pStreamRef = streamRef;

  return stemErr;

} /* aStemInternal_CreateTEAFileOutput */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * sStemTEAFile_Put
 */

aErr sStemTEAFile_Put(char* pData,
	   	      void* ref)
{
  aErr stemErr = aErrNone;
  aTEAFileState* pPrivate = (aTEAFileState*)ref;
  int nPreCt;
  int* pBuffCt;
  aPacketRef outPacket;
  int nLength;


  if (pPrivate == NULL)
    stemErr = aErrIO;
  
  if (stemErr == aErrNone)
    pBuffCt = &(pPrivate->nBytesInBuffer);

  if (stemErr == aErrNone) {

    /* update checksum with outgoing character */
    pPrivate->checksum += (unsigned char)(*pData);
    
    /* stick byte in buffer, +1 offset for command */
    pPrivate->packetBuffer[(*pBuffCt) + 1] = (*pData);
    (*pBuffCt)++;

    /* +2 offset for file size in EEPROM */
    nPreCt = (*pBuffCt) + (pPrivate->nBytesSent) + 2;

    /* if buffer full or EEPROM page boundary reached then 
     * ship packet */
    if (((*pBuffCt) == 7) || ((nPreCt % 64) == 0)) {

      /* send cmdTEADATA packet */
      nLength = (*pBuffCt) + 1;
      pPrivate->packetBuffer[0] = cmdTEADATA;
      stemErr = aPacketInternal_Create(pPrivate->pStem, 
        			       pPrivate->address, 
        			       (unsigned char) nLength,
                            	       pPrivate->packetBuffer, 
                            	       (aPacket**)&outPacket);
      if (stemErr == aErrNone) {

        aErr drainErr;
        /* drain the input stream since this may be a tight loop 
         * calling us and we don't want to starve the heartbeat
         */
        drainErr = aStemInternal_DrainStream(pPrivate->pStem,
      					pPrivate->pStem->linkStream,
      					&pPrivate->pStem->curLinkPacket,
      					NULL, NULL,
      					aTrue,
      					aPacketFromStem,
      					pPrivate->pStem->pLinkPacketList);
        aAssert(drainErr == aErrNone);
      
        stemErr = aStemInternal_SendPacket(pPrivate->pStem, 
          				   outPacket);
      }

      if (stemErr == aErrNone) {
        pPrivate->nBytesSent += (*pBuffCt);
        (*pBuffCt) = 0;
      }
    }
  }

  return(stemErr);

} /* sStemTEAFile_Put */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * sStemTEAFile_Get
 */

static aErr sStemTEAFile_Get(char* pData,
			      void* ref)
{
  aErr stemErr = aErrNone;
  aTEAFileState* pPrivate = (aTEAFileState*)ref;

  int* pBuffCt;
  aPacketRef outPacket;
  aPacketRef inPacket;

  unsigned char ucLength;
  unsigned char ucAddress;
  char data[aSTEMMAXPACKETBYTES];
  char datum;

  int ii;
  int nDataSize;
  
  if (pPrivate == NULL)
    stemErr = aErrIO;
  
  
  if ((stemErr == aErrNone) && (pPrivate->bDone == aTrue))
    stemErr = aErrEOF;

  pBuffCt = &(pPrivate->nBytesInBuffer);
    
  if (stemErr == aErrNone) {  

    /* check for empty buffer */
    if (*pBuffCt == 0) {

      /* send cmdTEAREAD command */
      ucLength = 1;
      data[0] = cmdTEAREAD;
      stemErr = aPacketInternal_Create(pPrivate->pStem, 
      			       	       pPrivate->address, ucLength,
                               	       data, (aPacket**)&outPacket);
      if (stemErr == aErrNone)
        stemErr = aStemInternal_SendPacket(pPrivate->pStem, outPacket);

      /* seek cmdTEADATA packet reply */
      if (stemErr == aErrNone) {
        stemErr = aStemInternal_GetPacket(pPrivate->pStem, 
        				  (aPacketFilter)sDataFilter,
        				  NULL,
        				  aTEAFILETIMEOUTMS,
        				  (aPacket**)&inPacket);
        if (stemErr == aErrNone)
          stemErr = aPacketInternal_GetData(pPrivate->pStem,
          			            inPacket,
          			            &ucAddress,
          			            &ucLength,
          			            data);
        if (stemErr == aErrNone) {
        
          if (1) {
            aPacket* pTemp = pPrivate->pStem->pLinkPacketList->pHead;
            char formatted[30];
            aStemInternal_DebugLine(pPrivate->pStem, "packetstart");
            while (pTemp != NULL) {
              aPacketInternal_Format(pTemp, formatted, 30);
              aStemInternal_DebugLine(pPrivate->pStem, formatted);
              pTemp = pTemp->pNext;
            }
            aStemInternal_DebugLine(pPrivate->pStem, "packetend");
          }

          switch (data[0]) {
  
          case cmdTEADATA:
            nDataSize = ucLength - 1;
            for (ii = 0; ii < nDataSize; ii++) {
              datum = data[nDataSize-ii];
              pPrivate->packetBuffer[ii] = datum;
              pPrivate->checksum += (unsigned char)datum;
            }
            *pBuffCt = nDataSize;
            pPrivate->nBytesSent += nDataSize;
            break;

          case cmdTEAEOF:
            pPrivate->bDone = aTrue;
            break;

          } /* switch */

          stemErr = aPacketInternal_Destroy(pPrivate->pStem, 
            				    inPacket);
        }
      }
    } /* buf count = 0 */
  } /* stemErr == aErrNone */

  /* if we are done, report the end of file condition */
  if (pPrivate->bDone == aTrue) {
    stemErr = aErrEOF;
  } else {
  
    /* else send one character from buffer */
    if ((stemErr == aErrNone) && (*pBuffCt > 0)) {
      (*pBuffCt)--;
      *pData = pPrivate->packetBuffer[(*pBuffCt)];
    }
  }

  return(stemErr);

} /* sStemTEAFile_Get */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * sStemTEAFile_DeleteInput
 */

aErr sStemTEAFile_DeleteInput(void* ref)
{
  aErr stemErr = aErrNone;
  aTEAFileState* pPrivate = (aTEAFileState*)ref;
  unsigned char ucLength;
  unsigned char ucAddress;
  unsigned char ucChecksum;
  char data[aSTEMMAXPACKETBYTES];

  aPacketRef inPacket;
  aPacketRef outPacket;

  if (pPrivate == NULL)
    stemErr = aErrIO;

  /* send cmdTEASUM */
  if (stemErr == aErrNone) {
    ucLength = 1;
    pPrivate->packetBuffer[0] = cmdTEASUM;
    stemErr = aPacketInternal_Create(pPrivate->pStem, 
    				     pPrivate->address, 
    				     ucLength,
                        	     pPrivate->packetBuffer, 
                        	     (aPacket**)&outPacket);
  }
  if (stemErr == aErrNone)
      stemErr = aStemInternal_SendPacket(pPrivate->pStem, 
      					 outPacket);

  /* get the checksum reply */
  if (stemErr == aErrNone)
    stemErr = aStemInternal_GetPacket(pPrivate->pStem, 
      				      (aPacketFilter)sCheckSumFilter,
      				      NULL,
      				      aTEAFILETIMEOUTMS,
      				      (aPacket**)&inPacket);
  if (stemErr == aErrNone)
    stemErr = aPacketInternal_GetData(pPrivate->pStem, 
        			      inPacket, 
        			      &ucAddress, 
                               	      &ucLength, 
                               	      data);

  if (stemErr == aErrNone)
    stemErr = aPacketInternal_Destroy(pPrivate->pStem, inPacket);
  
  if (stemErr == aErrNone) {
    ucChecksum = (unsigned char)data[1];
    if (ucChecksum != pPrivate->checksum) 
      stemErr = aErrRead;
  }

  /* send cmdTEAEOF */
  if (stemErr == aErrNone) {
    ucLength = 1;
    pPrivate->packetBuffer[0] = cmdTEAEOF;
    stemErr = aPacketInternal_Create(pPrivate->pStem, 
      				     pPrivate->address, 
      				     ucLength,
                          	     pPrivate->packetBuffer, 
                          	     (aPacket**)&outPacket);
    if (stemErr == aErrNone)
      stemErr = aStemInternal_SendPacket(pPrivate->pStem, 
        				 outPacket);                  	 
  }

  /* clean up the private stream data */
  if (pPrivate != NULL)
    aMemFree((aMemPtr)pPrivate);

  return(stemErr);

} /* sStemTEAFile_DeleteInput */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * sStemTEAFile_DeleteOutput
 */

aErr sStemTEAFile_DeleteOutput(void* ref)
{
  aErr stemErr = aErrNone;
  aTEAFileState* pPrivate = (aTEAFileState*)ref;
  int* pBuffCt;
  unsigned char ucLength;
  unsigned char ucAddress;
  unsigned char ucChecksum;
  char data[aSTEMMAXPACKETBYTES];

  aPacketRef inPacket = NULL;
  aPacketRef outPacket;

  if (pPrivate == NULL)
    stemErr = aErrIO;

  if (stemErr == aErrNone) {

    /* ship any remaining bytes */
    pBuffCt = &(pPrivate->nBytesInBuffer);

    if ((*pBuffCt) > 0) {

      /* send cmdTEADATA packet */
      ucLength = (unsigned char)((*pBuffCt)+1);
      pPrivate->packetBuffer[0] = cmdTEADATA;
      stemErr = aPacketInternal_Create(pPrivate->pStem, 
      				       pPrivate->address, 
      				       ucLength,
                          	       pPrivate->packetBuffer, 
                          	       (aPacket**)&outPacket);
      if (stemErr == aErrNone)
        stemErr = aStemInternal_SendPacket(pPrivate->pStem, 
        			           outPacket);
      if (stemErr == aErrNone) {
        pPrivate->nBytesSent += (*pBuffCt);
        (*pBuffCt) = 0;
      }
    }
  }

  /* send cmdTEASUM to get the checksum */
  if (stemErr == aErrNone) {
    ucLength = 1;
    pPrivate->packetBuffer[0] = cmdTEASUM;
    stemErr = aPacketInternal_Create(pPrivate->pStem, 
      				     pPrivate->address, 
      				     ucLength,
                          	     pPrivate->packetBuffer, 
                          	     (aPacket**)&outPacket);
    if (stemErr == aErrNone)
      stemErr = aStemInternal_SendPacket(pPrivate->pStem, 
        				 outPacket);
  }

  /* get the checksum reply */
  if (stemErr == aErrNone)
    stemErr = aStemInternal_GetPacket(pPrivate->pStem, 
      				      (aPacketFilter)sCheckSumFilter,
      				      NULL,
      				      aTEAFILETIMEOUTMS,
      				      (aPacket**)&inPacket);
  if (stemErr == aErrNone)
    stemErr = aPacketInternal_GetData(pPrivate->pStem, 
        			      inPacket, 
        			      &ucAddress, 
                               	      &ucLength, 
                               	      data);

  if (stemErr == aErrNone)
    stemErr = aPacketInternal_Destroy(pPrivate->pStem,
    			    	      inPacket);
  
  if (stemErr == aErrNone) {
    ucChecksum = (unsigned char)data[1];
    if (ucChecksum != pPrivate->checksum) 
      stemErr = aErrWrite;
  }

  /* send cmdTEAEOF */
  if (stemErr == aErrNone) {
    ucLength = 1;
    pPrivate->packetBuffer[0] = cmdTEAEOF;
    stemErr = aPacketInternal_Create(pPrivate->pStem, 
      				     pPrivate->address, 
      				     ucLength,
                          	     pPrivate->packetBuffer, 
                          	     (aPacket**)&outPacket);
    if (stemErr == aErrNone)
      stemErr = aStemInternal_SendPacket(pPrivate->pStem, 
        				 outPacket);                  	 
  }

  /* clean up the private stream data */
  if (pPrivate != NULL)
    aMemFree((aMemPtr)pPrivate);

  return stemErr;

} /* sStemTEAFile_DeleteOutput */



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * sInitFilter
 */

aBool sInitFilter(const unsigned char address,
		  const unsigned char length,
		  const char* data)
{
  if (data[0] == cmdTEAINIT)
    return aTrue;

  return aFalse;

} /* sInitFilter */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * sCheckSumFilter
 */

aBool sCheckSumFilter(const unsigned char address,
		              const unsigned char length,
		              const char* data)
{
  if (data[0] == cmdTEASUM)
    return aTrue;

  return aFalse;

} /* sCheckSumFilter */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * sDataFilter
 */

aBool sDataFilter(const unsigned char address,
		  const unsigned char length,
		  const char* data)
{  
  if ((data[0] == cmdTEAEOF) ||
      (data[0] == cmdTEADATA))
    return aTrue;
    
  return aFalse;

} /* sDataFilter */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * sStemTEAFile_Init
 */

aErr sStemTEAFile_Init(aTEAFileState* pFileState)
{
  aErr stemErr = aErrNone;
  aPacketRef inPacket = NULL;
  aPacketRef outPacket = NULL;
  unsigned char ucLength;
  unsigned char ucAddress;
  char data[aSTEMMAXPACKETBYTES];

  /* send cmdTEAINIT */
  if (stemErr == aErrNone) {

    unsigned char length;
    char data[aSTEMMAXPACKETBYTES];

    length = 2;
    data[0] = cmdTEAINIT;
    data[1] = (char)pFileState->nFileNumber;
    stemErr = aPacketInternal_Create(pFileState->pStem, 
    				     pFileState->address, 
    				     length, 
    				     data,
 				     (aPacket**)&outPacket);
    if (stemErr == aErrNone)
      stemErr = aStemInternal_SendPacket(pFileState->pStem, 
      					 outPacket);
  }

  /* wait for init reply */
  if (stemErr == aErrNone)
    stemErr = aStemInternal_GetPacket(pFileState->pStem, 
      				      (aPacketFilter)sInitFilter,
      				      NULL,
      				      aTEAFILETIMEOUTMS,
      				      (aPacket**)&inPacket);
  if (stemErr == aErrNone)
    stemErr = aPacketInternal_GetData(pFileState->pStem, 
        			      inPacket, 
        			      &ucAddress, 
                               	      &ucLength, 
                               	      data);

  if (stemErr == aErrNone)
    stemErr = aPacketInternal_Destroy(pFileState->pStem,
    			    	      inPacket);

  /* validate the init reply */
  if (stemErr == aErrNone) {
    /* packet data should be init, could be file error */
  }

  return stemErr;

} /* sStemTEAFile_Init */
