/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                 */
/* file: aStream.c                                                 */
/*                                                                 */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                 */
/* description: Cross-Platform implementation of stream I/O        */
/*              routines.                                          */
/*                                                                 */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                 */
/* Copyright  1997-2000. Acroname Inc.                            */
/*                                                                 */
/* This software is the property of Acroname Inc.  Any             */
/* distribution, sale, transmission, or re-use of this code is     */
/* strictly forbidden except with permission from Acroname Inc.    */
/*                                                                 */
/* To the full extent allowed by law, Acroname Inc. also excludes  */
/* for itself and its suppliers any liability, wheither based in   */
/* contract or tort (including negligence), for direct,            */
/* incidental, consequential, indirect, special, or punitive       */
/* damages of any kind, or for loss of revenue or profits, loss of */
/* business, loss of information or data, or other financial loss  */
/* arising out of or in connection with this software, even if     */
/* Acroname Inc. has been advised of the possibility of such       */
/* damages.                                                        */
/*                                                                 */
/* Acroname Inc.                                                   */
/* www.acroname.com                                                */
/* 720-564-0373                                                    */
/*                                                                 */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "aDefs.h"
#include "aOSDefs.h"
#include "aIO.h"
#include "aStream.h"


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * local routines for look ahead buffer used by stream read routines
 */

static aBool sStreamGetBufferedChar(aStream* pStream, 
				    char* pCharacter);
static aErr sStreamBufferChar(aStream* pStream, 
			      char character);


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * defines for file streams
 */
typedef struct aFileStreamData {
  aIOLib    ioRef;
  aFileRef  fileRef;
} aFileStreamData;


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * local static file stream callbacks
 */

static aErr sFileStreamGet(char* pData, void* ref);
static aErr sFileStreamPut(char* pData, void* ref);
static aErr sFileStreamWrite(const char* pData, 
		             const unsigned long nSize,
		             void* ref);
static aErr sFileStreamDelete(void* ref);


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * sStreamGetBufferedChar
 */

aBool sStreamGetBufferedChar(aStream* pStream, 
			     char* pCharacter)
{
  if (pStream->bufCount == 0)
    return aFalse;

  /* retrieve the character from the buffer */
  *pCharacter = pStream->pBuffer[pStream->bufFirst];
  pStream->bufFirst = (pStream->bufFirst + 1) % pStream->bufSize;
  pStream->bufCount--;

  return aTrue;

} /* sStreamGetBufferedChar */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * sStreamBufferChar
 */

aErr sStreamBufferChar(aStream* pStream, 
		       char character)
{
  aErr ioErr = aErrNone;
  
  aAssert(pStream);

  /* if no buffer is present, build one */
  if (pStream->pBuffer == NULL) {
    pStream->pBuffer = aMemAlloc(aSTREAMBUFCHUNK);
    if (pStream->pBuffer == NULL)
      ioErr = aErrMemory;
    else {
      pStream->bufSize = aSTREAMBUFCHUNK;
      pStream->bufCount = 0;
      pStream->bufLast = 0;
      pStream->bufFirst = 0;
    }
  }

  aAssert(pStream->pBuffer);

  /* now, see if the buffer is full */
  if ((ioErr == aErrNone)
      && (pStream->bufCount >= pStream->bufSize - 1)) {
    /* grow the buffer */
    char* pNew = aMemAlloc((aMemSize)(pStream->bufSize + aSTREAMBUFCHUNK));
    if (pNew == NULL)
      ioErr = aErrMemory;
    else {
      unsigned int c1Size, c2Size;

      /* copy from the first to the end of the existing */
      c1Size = ((pStream->bufLast > pStream->bufFirst) ?
                	pStream->bufLast : 
                	pStream->bufSize) - pStream->bufFirst;
      aMemCopy(pNew, &pStream->pBuffer[pStream->bufFirst], c1Size);
      
      /* now, handle any wrap */
      if (pStream->bufLast < pStream->bufFirst) {
        c2Size = pStream->bufLast;
        if (c2Size)
          aMemCopy(&pNew[c1Size], pStream->pBuffer, c2Size);
      } else
        c2Size = 0;
      
      /* now, switch out the smaller buffer for the larger copy */
      pStream->bufFirst = 0;
      pStream->bufLast = c1Size + c2Size;
      aAssert(pStream->bufCount == pStream->bufLast);
      pStream->bufSize = pStream->bufSize + aSTREAMBUFCHUNK;
      aMemFree(pStream->pBuffer);
      pStream->pBuffer = pNew;
    }
  }

  /* add the character to the buffer */
  if (ioErr == aErrNone) {
    aAssert(pStream->pBuffer);
    pStream->pBuffer[pStream->bufLast] = character;
    pStream->bufLast = (pStream->bufLast + 1) % pStream->bufSize;
    pStream->bufCount++;
  }

  return ioErr;

} /* sStreamBufferChar */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aFileStreamGet
 */

aErr sFileStreamGet(char* pData,
		    void* ref)
{
  aErr ioErr = aErrNone;
  aFileStreamData* data = (aFileStreamData*)ref;

  if (data == NULL)
    ioErr = aErrIO;
  else
    aFile_Read(data->ioRef, data->fileRef, pData, 1, NULL, &ioErr);

  return(ioErr);

} /* end of sFileStreamGet */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * sFileStreamPut
 */

aErr sFileStreamPut(char* pData,
		    void* ref)
{
  aErr ioErr = aErrNone;
  aFileStreamData* data = (aFileStreamData*)ref;

  if (data == NULL)
    ioErr = aErrIO;
  else {
    aFile_Write(data->ioRef, data->fileRef, pData, 1, NULL, &ioErr);
  }

  return(ioErr);

} /* end of sFileStreamPut */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * sFileStreamWrite
 */

aErr sFileStreamWrite(const char* pData, 
		      const unsigned long nSize,
		      void* ref)
{
  aErr ioErr = aErrNone;
  aFileStreamData* data = (aFileStreamData*)ref;

  /* check for valid stream data */
  aFile_Write(data->ioRef, data->fileRef, pData, nSize, NULL, &ioErr);

  return ioErr;

} /* sFileStreamWrite */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aFileStreamDelete
 */

aErr sFileStreamDelete(void* ref)
{
  aErr ioErr = aErrNone;
  aFileStreamData* data = (aFileStreamData*)ref;

  if (data == NULL)
    ioErr = aErrIO;
  else {
    aFile_Close(data->ioRef, data->fileRef, &ioErr);
    aMemFree((aMemPtr)data);
  }

  return(ioErr);

} /* end of sFileStreamDelete */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aStream_Create
 */

aIO_EXPORT aLIBRETURN aStream_Create(aIOLib ioRef,
				     aStreamGetProc getProc,
				     aStreamPutProc putProc,
				     aStreamDeleteProc deleteProc,
				     const void* procRef,
				     aStreamRef* pStreamRef,
				     aErr* pErr)
{
  aErr ioErr = aErrNone;
  aStream* pStream;

  /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   * check params and initialize
   */
  if (pStreamRef == NULL)
    ioErr = aErrParam;
  else
    *pStreamRef = NULL;

  if (ioErr == aErrNone) {
    pStream = (aStream*)aMemAlloc(sizeof(aStream));
    if (pStream == NULL) {
      ioErr = aErrMemory;
    } else {
     aBZero(pStream, sizeof(aStream));
     pStream->getProc = getProc;
     pStream->putProc = putProc;
     pStream->deleteProc = deleteProc;
     pStream->procRef = (void*)procRef;
     pStream->check = aSTREAM_CHECK;
     pStream->libRef = ioRef;
     *pStreamRef = pStream;
    }
  }

  if (pErr != NULL)
    *pErr = ioErr;

  return((aLIBRETURN)(ioErr != aErrNone));

} /* end of aStream_Create routine */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aStream_CreateFileInput
 */

aIO_EXPORT aLIBRETURN aStream_CreateFileInput(aIOLib ioRef, 
					      const char *pFilename,
				 	      const aFileArea eArea,
					      aStreamRef* pStreamRef,
					      aErr* pErr)
{
  aFileRef inputFile = NULL;
  aStreamRef streamRef = NULL;
  aFileStreamData* fileData;
  aErr ioErr = aErrNone;

  /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   * check params
   */
  if ((pFilename == NULL) || (pStreamRef == NULL))
    ioErr = aErrParam;

  /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   * try to open the file
   */
  if (ioErr == aErrNone)
    aFile_Open(ioRef, pFilename, aFileModeReadOnly, eArea,
    	       &inputFile, &ioErr);

  /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   * create the private stream data
   */
  if (ioErr == aErrNone) {
    fileData = (aFileStreamData*)aMemAlloc(sizeof(aFileStreamData));
    if (fileData == NULL)
      ioErr = aErrMemory;
    else {
      fileData->ioRef = ioRef;
      fileData->fileRef = inputFile;
    }
  }

  /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   * create the stream
   */
  if (ioErr == aErrNone)
    aStream_Create(ioRef, 
    		   sFileStreamGet,
    		   NULL,
    		   sFileStreamDelete,
    		   fileData,
    		   &streamRef, 
    		   &ioErr);

  if (ioErr == aErrNone)
    *pStreamRef = streamRef;
  else {
    if (inputFile != NULL) {
      aFile_Close(ioRef, inputFile, NULL);
    }
  }

  if (pErr != NULL)
    *pErr = ioErr;

  return (aLIBRETURN)(ioErr != aErrNone);
  
} /* aStream_CreateFileInput */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aStream_CreateFileOutput
 */

aIO_EXPORT aLIBRETURN aStream_CreateFileOutput(aIOLib ioRef,
					       const char *pFilename,
					       const aFileArea eArea,
					       aStreamRef* pStreamRef,
					       aErr* pErr)
{
  aErr ioErr = aErrNone;
  aStreamRef streamRef;
  aFileRef outputFile;
  aFileStreamData* fileData;

  /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   * check params
   */
  if ((pFilename == NULL) || (pStreamRef == NULL))
    ioErr = aErrParam;

  /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   * try to open the file
   */
  if (ioErr == aErrNone) {
    aFile_Open(ioRef, pFilename, aFileModeWriteOnly, eArea,
    	       &outputFile, &ioErr);
  }

  /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   * create the private file stream data
   */
  if (ioErr == aErrNone) {
    fileData = (aFileStreamData*)aMemAlloc(sizeof(aFileStreamData));
    if (fileData == NULL)
      ioErr = aErrMemory;
    else {
      fileData->ioRef = ioRef;
      fileData->fileRef = outputFile;
    }
  }

  /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   * build the actual stream
   */
  if (ioErr == aErrNone) {
    aStream_Create(ioRef, 
    		   NULL,
    		   sFileStreamPut,
    		   sFileStreamDelete,
    		   fileData,
    		   &streamRef, 
    		   &ioErr);
    if (ioErr != aErrNone)
      aFile_Close(ioRef, outputFile, NULL);
  }

  if (ioErr == aErrNone) {
    aStream* pStream = (aStream*)streamRef;
    pStream->writeProc = sFileStreamWrite;
    *pStreamRef = streamRef;
  }
  
  if (pErr != NULL)
    *pErr = ioErr;

  return (aLIBRETURN)(ioErr != aErrNone);

} /* end of aStream_CreateFileOutput */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aStream_Read
 */

aIO_EXPORT aLIBRETURN aStream_Read(aIOLib ioRef, 
				   aStreamRef streamRef,
				   char* pBuffer,
    				   const unsigned long ulLength,
    				   aErr* pErr)
{
  unsigned long i;
  aErr ioErr = aErrNone;
  char* p = pBuffer;
  aStream* pStream = (aStream*)streamRef;
  
  /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   * check params
   */
  aSTREAMCHECK(pStream);
  if (pBuffer == NULL)
    ioErr = aErrParam;
  if ((ioErr == aErrNone) && (pStream->getProc == NULL))
    ioErr = aErrMode;

  /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   * do the read
   */
  for (i = 0; (ioErr == aErrNone) && (i < ulLength); i++, p++)
    ioErr = pStream->getProc(p, pStream->procRef);

  if (pErr != NULL)
    *pErr = ioErr;

  return (aLIBRETURN)(ioErr != aErrNone);

} /* end of aStream_Read */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aStream_Write
 */

aIO_EXPORT aLIBRETURN aStream_Write(aIOLib ioRef, 
				    aStreamRef streamRef,
				    const char* pBuffer,
    				    const unsigned long ulLength,
    				    aErr* pErr)
{
  aErr ioErr = aErrNone;
  unsigned long i;
  char *p;
  aStream* pStream = (aStream*)streamRef;


  /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   * check params
   */
  aSTREAMCHECK(pStream);
  aAssert(ioErr == aErrNone);
  if (pBuffer == NULL)
    return(aErrParam);
  if (pStream->putProc == NULL) {
    return(aErrMode);
  }

  if (pStream->writeProc != NULL) {
    ioErr = pStream->writeProc(pBuffer, ulLength, pStream->procRef);
  } else {
    p = (char *)pBuffer;
    for (i = 0; (ioErr == aErrNone) && (i < ulLength); i++, p++) 
      ioErr = pStream->putProc(p, pStream->procRef);
  }

  if (pErr != NULL)
    *pErr = ioErr;

  return (aLIBRETURN)(ioErr != aErrNone);

} /* aStream_Write */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aStream_ReadLine
 *
 *  For the record:
 *   Mac text lines end with 0x0D ("\r")
 *   Unix text lines end with 0x0A ("\n")
 *   DOS text lines end with 0x0D 0x0A ("\r\n")
 */

aIO_EXPORT aLIBRETURN aStream_ReadLine(aIOLib ioRef, 
				       aStreamRef streamRef,
				       char* pBuffer,
    				       const unsigned long ulLength,
    				       aErr* pErr)
{
  aErr ioErr = aErrNone;
  aErr copyErr;
  unsigned long ulCount = 0;
  char in, ahead;
  aBool bDone = aFalse;
  aStream* pStream = (aStream*)streamRef;
  unsigned int i;


  /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   * check params
   */
  aSTREAMCHECK(pStream);
  if ((pBuffer == NULL) || (ioRef == NULL))
    ioErr = aErrParam;

  /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   * need at least one byte for null termination
   */
  if ((ioErr == aErrNone) && (ulLength < 1))
    ioErr = aErrParam;

  /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   * get characters until we reach EOF, not ready, or end of line
   */
  while ((ioErr == aErrNone)
         && (bDone == aFalse)) {

    /* get the next character */
    if (!sStreamGetBufferedChar(pStream, &in))
      aStream_Read(ioRef, streamRef, &in, 1, &ioErr);

    switch (ioErr) {

    /* EOF means we are done */
    case aErrEOF:
      /* the last line could end in EOF */
      if (ulCount != 0) {
        ioErr = aErrNone;
        bDone = aTrue;
      }
      break;

    /* not ready means the data isn't here yet */
    case aErrNotReady:
      /* buffer what we have so far */
      copyErr = aErrNone;
      for (i = 0; (copyErr == aErrNone) && (i < ulCount); i++)
        copyErr = sStreamBufferChar(pStream, pBuffer[i]);
      break;
    
    case aErrNone:
      switch (in) {

      case '\n':
        /* eat the newline and bail */
        bDone = aTrue; 
        break;

      case '\r':
        /* look ahead at the next character */
        if (!sStreamGetBufferedChar(pStream, &ahead))
          aStream_Read(ioRef, streamRef, &ahead, 1, &ioErr);

        if (ioErr == aErrNone) {

          /* gobble the \r\n and we are done */
          if (ahead == '\n') {
            bDone = aTrue;
            break;
          }
          /* next char seems valid start of line, we are done */
          bDone = aTrue;
          ioErr = sStreamBufferChar(pStream, ahead);

        /* special case for files that end in an \r */
        } else if (ioErr == aErrEOF) {
          ioErr = aErrNone;
          bDone = aTrue;
        }
        break;

      default:
        pBuffer[ulCount++] = in;
        break;

      } /* end switch */
      break;

    default:
      break;
 
    } /* switch */
  }

  if (ioErr == aErrNone)
    pBuffer[ulCount] = '\0';   /* null terminate */

  if (pErr != NULL)
    *pErr = ioErr;

  return (aLIBRETURN)(ioErr != aErrNone);

} /* end of aStream_ReadLine */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aStream_WriteLine
 */

aIO_EXPORT aLIBRETURN aStream_WriteLine(aIOLib ioRef, 
					aStreamRef streamRef,
					const char* pBuffer,
				        aErr* pErr)
{
  aErr ioErr = aErrNone;

  /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   * check params
   */
  if ((pBuffer == NULL) || (ioRef == NULL))
    ioErr = aErrParam;
  aSTREAMCHECK(streamRef);

  /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   * write the line
   */
  if (ioErr == aErrNone)
    aStream_Write(ioRef, streamRef, pBuffer, 
  		  aStringLen(pBuffer), &ioErr);

  /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
   * write line terminator (aszEOL is OS-specific, see OSDefs.h)
   */
  if (ioErr == aErrNone)
    aStream_Write(ioRef, streamRef, aszEOL, 
  		  aStringLen(aszEOL), &ioErr);

  if (pErr != NULL)
    *pErr = ioErr;

  return (aLIBRETURN)(ioErr != aErrNone);

} /* aStream_WriteLine */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aStream_Destroy
 */

aIO_EXPORT aLIBRETURN aStream_Destroy(aIOLib ioRef,
				      aStreamRef streamRef,
				      aErr* pErr)
{
  aErr ioErr = aErrNone;
  aStream* pStream = (aStream*)streamRef;

  /* check params */
  aSTREAMCHECK(pStream);

  if ((ioErr == aErrNone) && (pStream->deleteProc != NULL))
    ioErr = pStream->deleteProc(pStream->procRef);

  /* clean up any buffer that may have been allocated */
  if ((ioErr == aErrNone)
      && (pStream->pBuffer != NULL)) {
    aMemFree(pStream->pBuffer);
    pStream->pBuffer = NULL;
  }

  /* clean up the actual stream object */
  if (ioErr == aErrNone) {
    pStream->check = 0; /* invalidate the check */
    aMemFree((aMemPtr)pStream);
  }

  if (pErr != NULL)
    *pErr = ioErr;

  return (aLIBRETURN)(ioErr != aErrNone);

} /* end of aStream_Destroy */
