/*
  Kalid Azad - Senior Independent Work - brainstem.c
  Functions to talk to the brainstem
 */

#include "brainstem.h"
#include "debug.h"

int sendPacket(robotConfig *r, char len, char *data);

/* Setup libraries, serial streams for brainstem */
int stem_open(robotConfig *r){
  aErr err = aErrNone;
  int i;

  aBZero(r, sizeof(robotConfig) );

  r->module = DEFAULTADDRESS; /* typically address 2 */

  /* open library aIO */
  if (aIO_GetLibRef(&(r->ioRef), &err) ){
    printf("Error in aIO_GetLibRef\n");
    return -1;
  }
  if (err != aErrNone)
    return -1;

  /* output stream */
  err = aStream_Create_STDIO_Console_Output(r->ioRef, &(r->outputStream) );
  if (err != aErrNone){
    printf("Error in aStram_Create_STDIO_Console_Output\n");
    return -1;
  }

  /* open Stem library */
  if (aStem_GetLibRef(&(r->stemRef), NULL)){
    printf("Error in aStem_GetLibRef\n");
    return -1;
  }

  /* stream connection to serial port */
  if (aStream_CreateSerial(r->ioRef, PORTNAME, BAUDRATE, &(r->serialStream), NULL) ){
    printf("Error in aStream_CreateSerial");
    return -1;
  }

  /* set serial stream for stem library */
  if (aStem_SetStream(r->stemRef, r->serialStream, kStemModuleStream, NULL) ){
    printf("Error in aStem_SetStream");
    return -1;
  }

  /* create packet for the connection */
  r->len = 10; /* aMAXPACKETSIZE */
  for (i = 0; i < r->len; i++) 
    r->data[i] = 0;

  /*
  aPacket_Create(r->stemRef, r->module, r->len, r->data, &(r->packet), &err);
  if (err != aErrNone){
    printf("Error in aPacket_Create\n");
    return -1;
  }
  */

  return aErrNone; /* no errors */
}

/* shutdown connection, do cleanup */
int stem_close(robotConfig *r){
  aStem_ReleaseLibRef(r->stemRef, NULL);
  aStream_Destroy(r->ioRef, r->serialStream, NULL);
  aIO_ReleaseLibRef(r->ioRef, NULL);
  /* aLeakCheckCleanup(); */
}

/* 
   Send command with N arguments (CmdN is cmdName + N args) 
   Inserts len and creates array automatically 
*/
int sendCmd0(robotConfig *r, int cmdName){
  char data[1];
  data[0] = cmdName;
  return sendPacket(r, 1, data);
}

int sendCmd1(robotConfig *r, int cmdName, char arg1){
  char data[2];
  data[0] = cmdName;
  data[1] = arg1;
  return sendPacket(r, 2, data);
}

int sendCmd2(robotConfig *r, int cmdName, char arg1, char arg2){
  char data[3];
  data[0] = cmdName;
  data[1] = arg1;
  data[2] = arg2;
  return sendPacket(r, 3, data);
}

int sendCmd3(robotConfig *r, int cmdName, char arg1, char arg2, char arg3){
  char data[4];
  data[0] = cmdName;
  data[1] = arg1;
  data[2] = arg2;
  data[3] = arg3;
  return sendPacket(r, 4, data);
}

/* Set absolute servo position for certain servo */
int SRV_ABS(robotConfig *r, char serv, char abspos){
  return sendCmd2(r, cmdSRV_ABS, serv, abspos);
}

/* Returns current servo position for servo */
int SRV_POS(robotConfig *r, char serv){
  char len = 10;
  char data[10];

  sendCmd1(r, cmdSRV_ABS, serv);
  if (getPacketData(r, 1000, &len, data) == aErrNone){
    return (int) data[2]; /* 3rd byte is pos */
  }
  else{
    printf("Error in SRV_POS: getPacketData\n");
    return -1;
  }
}

/* set relative servo position */
int SRV_REL(robotConfig *r, char serv, char reloff, char reldir){
  return sendCmd3(r, cmdSRV_REL, serv, reloff, reldir);
} 

/* save current servo setup (speed, offset, etc.) */
int SRV_SAV(robotConfig *r){
  return sendCmd0(r, cmdSRV_SAV);
}

/* reset brainstem, returning servos to neutral pos */
int RESET(robotConfig *r){
  return sendCmd0(r, cmdRESET);
}

/* Create and send packet */
int sendPacket(robotConfig *r, char len, char *data){
  aPacketRef packet;
  aErr err = aErrNone;
  int i;

  //ENTER_FUNCTION;
  printf("Enter sendPacket\n");

  printf("Sending: addr: 2 len: %d data:", len);
  for(i = 0; i < len; i++)
    printf(" %d ", data[i] );
  printf("\n");
  
  aPacket_Create(r->stemRef, r->module, len, data, &packet, &err);
  if (err != aErrNone){
    printf("Error in aPacket_Create\n");
    return err;
  }
  
  /* used to be packet, not r->packet */
  aStem_SendPacket(r->stemRef, packet, &err);

  aPacket_Destroy(r->stemRef, packet, NULL);

  if (err != aErrNone){
    printf("Error in sendPacket\n");
  }

  //EXIT_FUNCTION;
  printf("Exit sendPacket\n");
  return err;
}

/*
  Returns packet data and length,
   make sure data > aSTEMMAXPACKETBYTES 
*/
int getPacketData(robotConfig *r, int timeout, unsigned char *len, char *data){
  aPacketRef packet;
  aErr err = aErrNone;
  unsigned char addr; /* can return later, if needed */
  int i;
   
  char mydata[10];

  /* timeout = 1000; */  

  printf("Enter getPacketData\n");

  /*** should probably have packet as part of r structure ****/
  
  aPacket_Create(r->stemRef, r->module, 4, data, &packet, &err);
  if (err != aErrNone){
    printf("Error in aPacket_Create\n");
    return err;
  }
  
  aStem_GetPacket(r->stemRef, NULL, NULL, timeout, &packet, &err);
  if (err != aErrNone){
    printf("Error in aStem_GetPacket:%d\n", err);
    return err;
  }
  
  aPacket_GetData(r->stemRef, packet, &addr, len, data, &err);
  if( err != aErrNone){
    printf("Error in aPacket_GetData\n");
    return err;
  }

  printf("Data received: ");
  for (i = 0; i < *len; i++){
    printf(" %d ", data[i]); 
  }
  printf("\n");

  /*aPacket_Destroy(r->stemRef, packet, NULL); */

  printf("Exit getPacketData\n\r");
 
  return err;
}








