/* Kalid Azad - Senior Independent work 
   Robot control program
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h> /* make sure to compile -lm to link math lib */
#include "raw.h"
#include "brainstem.h"
#include "debug.h"

/* CONTROLS */
#define LEFT 'a'    /* 4 cardinal directions of motion */
#define RIGHT 'd'
#define FORWARD 'w'
#define BACK 's'
#define CW 'p'      /* clocwise and CCW rotation */
#define CCW 'o'
#define STOP '0'
#define CONTINUE '='      /*if on, robot keeps doing what it's doing */
#define FASTER  '+'
#define SLOWER  '-'
#define INVALID -1  /* unrecognized command */
#define INFO 'i'   /* print status of robot */
#define EXIT 'y' /* command to exit */

#define M1_STOP 128 /* default rest setting for motor */
#define M2_STOP 128
#define M3_STOP 128
#define SPEED_DEFAULT 5
#define ROTATION_DEFAULT 5
#define MAX_REVS 1.0 /* maximum revolutions/sec */


#define CHECK(x,y) if(x == 0){ printf(y); exit(1);}

struct robotState{
  int m1; /* current motor settings */
  int m2;
  int m3;

  int m1_stop;
  int m2_stop;
  int m3_stop;

  int speed; /* speed step -- amount motor changed */
  int rotation; /* amount to change rotation each time */
  int cont; /* continue mode */

  robotConfig *r;
};

typedef struct robotState * State;

int aPPRK_go(char m1, char m2, char m3, State S);
int turn(char dir, State S);

/* print current settings dynamically */
void printintro(){
  
  printf("===============================================\n");
  printf(" Kalid's Robot control program                 \n");
  printf(" Controls:    forward(%c)                      \n", FORWARD);
  printf("      left(%c)           right(%c)             \n", LEFT, RIGHT);
  printf("                back(%c)                       \n", BACK);
  printf("                                               \n");
  printf(" Rotation:                                     \n");
  printf("   CCW:%c     CW:%c                            \n", CCW, CW);
  printf("                                               \n");
  printf(" Stop:%c    Continue mode:%c                   \n", STOP, CONTINUE);
  printf(" Faster:%c   Slower:%c                         \n", FASTER, SLOWER);
  printf(" Info: %c                                      \n", INFO);
  printf("                                               \n");
  printf(" End program:%c                                \n", EXIT);
  printf("===============================================\n");
}

/* set State fields to defaults */
void initState(State x){
  CHECK(x, "Error in initState");

  setStop(x); /* set stop defaults */
    
  x->m1 = x->m1_stop;
  x->m2 = x->m2_stop;
  x->m3 = x->m3_stop;
  
  x->speed = SPEED_DEFAULT;
  x->rotation = ROTATION_DEFAULT;
  x->cont = 0; // off by default
}

/* 
   Run first
   Creates and inits state, prints intro msg 
   Returns State for current session
*/
void shutdown(State x);
State init(){
  State x;

  x = calloc( sizeof(*x), 1 );

  if (ttysetup() < 0){
    printf("Error in ttysetup\n");
    shutdown(x);
    exit(0); /* get chars 1 at a time */
  }

  x->r = malloc( sizeof(*(x->r)) );
  if ( stem_open(x->r) < 0 ){
    printf("Error in connecting to brainstem\n");
    shutdown(x);
    exit(0);
  }

  initState(x);

  return x;
}

/* Query robot for initial config */
int setStop(State x){
  int m1, m2, m3;

  m1 = SRV_POS(x->r, 0);
  m2 = SRV_POS(x->r, 1);
  m3 = SRV_POS(x->r, 2);

  if ( (m1 < 0) || (m2 < 0) || (m3 < 0) ){
    printf("Error in setStop: m1<%d> m2<%d> m3<%d>\n", m1, m2, m3);
    return -1;
  }

  x->m1_stop = x->m1 = m1;
  x->m2_stop = x->m2 = m2;
  x->m3_stop = x->m3 = m3;

  return 0;
}


/* Cleanup, garbage collection */
void shutdown(State x){
  printf("Exiting...\n");

  if (x->r != NULL)
    stem_close(x->r);

  free(x->r);
  free(x);
  ttyreset(0);
}

/* parse input, execute command, redisplay */
int main(){
  int x = 0;
  State robot;

  printintro();
  robot = init();

  while ( (x = getCommand() ) != EXIT){
    if (x == INVALID)
      continue;

    execute(x, robot);
    printf("\n\r"); /* clear the line */
  }
  
}

/* Returns command pressed, if any */
int getCommand(){
  int c = ttygetchar();
  printf("got %c\n", c);

  switch(c){
  case LEFT: 
  case RIGHT: 
  case FORWARD: 
  case BACK: 
  case CW: 
  case CCW: 
  case STOP: 
  case CONTINUE: 
  case FASTER: 
  case SLOWER:
  case EXIT:
  case INFO:
    return c;
  default:
    return INVALID;
  }
}

int execute(int x, State S){
  switch(x){
  case LEFT:
  case RIGHT: 
  case FORWARD: 
  case BACK:
    move(x, S);
    break;
  case CW: 
  case CCW:
    turn(x, S);
    break;
  case STOP:
    stop(S);
    break;
  case CONTINUE:
    S->cont ^= 1; /* toggle with XOR */
    break;
  case FASTER: 
  case SLOWER:
    speed(x, S);
    break;
  case INFO:
    show_info(S);
    break;
  default:
    printf("Bad parameter passed to execute\n");
    return INVALID;
  }
}

int show_info(State S){
  printf("====== Robot info =============\n");
  printf("m1-m3:%d %d %d                 \n", S->m1, S->m2, S->m3);
  printf("stop values:%d %d %d           \n", S->m1_stop, S->m2_stop, S->m3_stop);
  printf("speed: %d   continue:%d        \n", S->speed, S->cont);
  printf("====== End info ===============\n");
}

int speed(int x, State S){
  return 1;
}

static struct robotState orig;
int move(int x, State S){
  int angle;

  memcpy(&orig, S, sizeof(*S));
 
  switch(x){
  case LEFT:
    angle = 180;
    break;
  case RIGHT:
    angle = 0;
    break;
  case FORWARD:
    angle = 90;
    break;
  case BACK:
    angle = 270;
    break;
  default:
    printf("Bad param passed to move\n");
    return -1;
  }

  /* if in continue mode, just set angle and return*/
  if (S->cont){
    vector_move(angle, 1.0, S);
  }
  else{
    /* set motors for MOTORTIME seconds */
    /* set motor for some ticks */
    vector_move(angle, 1.0, S);
    /* stop(S); */
  }

}

/*Set motors to move at angle with certain speed 
  speed from 0 (min) to 1.0 (max)
  Even though we _can_ move and rotate, don't allow it.

  Ugly function, needs to be cleaned up
*/
int vector_move(int angle, double speed, State S){
  int i;

  double radians;
  double omega[3]; /* omega 0 - 2, angular velocity */
  double b= 0.090, r = 0.020; /* radius of robot and wheel, in meters */

  /* matrix */
  double V[2];
  double max, t, m;

  double F0[2] = {-1.0, 0.0};
  double F1[2] = {.5, .866};
  double F2[2] = {.5, -.866};
  

  /* ******************************************** */
  /* skip all that stuff for now... just make it twitch!! */
  DPRINT("Making it twitch!\n");
  aPPRK_go(200, 200, 200, S);
  return 1;
  /* ******************************************* */
  
  /* if speed negative, make pos and reverse angle */
  if ( speed < 1.0){
    speed *= -1;
    angle -= 180;
  }

  if (speed > 1.0) /* upper limit 1.0 */
    speed = 1.0;

  angle %= 360; /* keep angles nice */
  radians = (double) angle / 57.3;
  
  /* set X and Y components */
  V[0] = cos(radians);
  V[1] = sin(radians);
  
  printf("moving at angle: %d, speed: %f", angle, speed);

  omega[0] = ( F0[0]*V[0] + F0[1]*V[1] ) / r;
  omega[1] = ( F1[0]*V[0] + F1[1]*V[1] ) / r;
  omega[2] = ( F2[0]*V[0] + F2[1]*V[1] ) / r;
  
  /* scale by speed (from 0 to 1) */
  return aPPRK_go(Vel_To_Value(omega[0]), Vel_To_Value(omega[1]),
		  Vel_To_Value(omega[2]), S );
}

/* Convert wheel's angular velocity (radians/sec) to servo value */
int Vel_To_Value(double vel){
  double x; /* revs per sec */
  
  /* n radians/sec = n/(2*PI) revolutions/sec */
  x = vel / (3.1415 * 2);

  /* limit x to MAX_REVS */
  if ( fabs(x) > MAX_REVS ){
    x = x > 0 ? MAX_REVS : (-1.0 * MAX_REVS);
  }
   
  return (int) (39.8053*x*x*x - 12.6083*x*x + 22.1197*x + 128.4262);
}


/*
  Set all 3 motor speeds
  Don't add rotational offset 
*/
int aPPRK_go(char m1, char m2, char m3, State S){
  aErr ret1;
  aErr ret2;
  aErr ret3;

  /* try to change speed, update S if successful */
  DPRINT("Enter aPPRK_go\n");

  ret1 = SRV_ABS(S->r, 0, m1);
  ret2 = SRV_ABS(S->r, 1, m2);;
  ret3 = SRV_ABS(S->r, 2, m3);

  if ( (ret1 != aErrNone) || (ret2 != aErrNone) || (ret3 != aErrNone) ){
    printf("Error in aPPRK_go: 1<%d> 2<%d> 3<%d>\n", ret1, ret2, ret3);
    return -1;
  }
  else{
    S->m1 = m1;
    S->m2 = m2;
    S->m3 = m3;
  }

  DPRINT("Exit aPPRK_go\n");
 
  return 0;
}

/* add only rotational offset r (based on speed) */
int turn(char dir, State S){
  int i = 0;
  int ret = 0;
  int rotation = 0;

  /* may have to switch this! */
  rotation = (dir == CCW ? 1 : 0 );

  /* eventually, check to see what the max rotation we can
     support is... but not now */

  for(i = 0; i <= 2; i++){
    if (SRV_REL(S->r, i, S->rotation, rotation) == aErrNone)
      ret = -1; 
  }

  return ret;
}

int stop(State S){
  return RESET(S->r);
}
