/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                 */
/* file: unix_aIO.c                                               */
/*                                                                 */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                 */
/* description: implementation of file I/O routines for unix.     */
/*                                                                 */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                 */
/* Copyright  1997-2002. Acroname Inc.                            */
/*                                                                 */
/* This software is the property of Acroname Inc.  Any             */
/* distribution, sale, transmission, or re-use of this code is     */
/* strictly forbidden except with permission from Acroname Inc.    */
/*                                                                 */
/* To the full extent allowed by law, Acroname Inc. also excludes  */
/* for itself and its suppliers any liability, wheither based in   */
/* contract or tort (including negligence), for direct,            */
/* incidental, consequential, indirect, special, or punitive       */
/* damages of any kind, or for loss of revenue or profits, loss of */
/* business, loss of information or data, or other financial loss  */
/* arising out of or in connection with this software, even if     */
/* Acroname Inc. has been advised of the possibility of such       */
/* damages.                                                        */
/*                                                                 */
/* Acroname Inc.                                                   */
/* www.acroname.com                                                */
/* 720-564-0373                                                    */
/*                                                                 */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifdef aUNIX

#include <sys/time.h>
#include <signal.h>
#include <unistd.h>

#include "unix_aIO.h"


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aIO_GetLibRef
 */

aIO_EXPORT aLIBRETURN aIO_GetLibRef(aIOLib* pRef,			
				    aErr* pErr)
{
  aErr err = aErrNone;
  aUnixIOLib* pLib = NULL;
  
  if (pRef == NULL)
    err = aErrParam;
  
  if (err == aErrNone) {
    pLib = (aUnixIOLib*)aMemAlloc(sizeof(aUnixIOLib));
    if (pLib == NULL)
      err = aErrMemory;
  }

  if (err == aErrNone)
    err = aIOInternal_Initialize((aIO*)pLib);
 
  if (err == aErrNone)
    *pRef = pLib;

  if (pErr != NULL)
    *pErr = err;

  return err;

} /* end of aIO_GetLibRef routine */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aIO_ReleaseLibRef
 */

aIO_EXPORT aLIBRETURN aIO_ReleaseLibRef(aIOLib libRef,
					aErr* pErr)
{
  aErr err = aErrNone;

  aVALIDIO(libRef);

  if (err == aErrNone) {
    aMemFree((aUnixIOLib*)libRef);
  }

  if (pErr != NULL)
    *pErr = err;
  
  return err;

} /* end of aIO_ReleaseLibRef routine */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aIO_GetMSTicks
 */

aLIBRETURN aIO_GetMSTicks(aIOLib ioRef,
			  unsigned long* pNTicks,
			  aErr* pErr)
{
  aErr err = aErrNone;
  struct timeval ts;
  struct timezone tz;

  aVALIDIO(ioRef);

  if ((err == aErrNone) && (pNTicks == NULL))
    err = aErrParam;

  if (err == aErrNone) {
    if (gettimeofday(&ts, &tz)) {
      err = aErrIO;
    } else {
      /* here we scale down the seconds by 10000 to keep within a
       * long */
      *pNTicks = (unsigned long)
      	((ts.tv_sec % 10000) * 1000L + ts.tv_usec / 1000L);
    }
  }

  if (pErr != NULL)
    *pErr = err;

  return (aLIBRETURN)(err != aErrNone);

} /* aIO_GetMSTicks */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * aIO_MSSleep
 */

#if 0
/* signal handler for the timer wakeup (SIGALRM) */
static void alrm(int toss);
void alrm(int toss) {
}
#endif

aLIBRETURN aIO_MSSleep(aIOLib ioRef,
		       const unsigned long msTime,
		       aErr* pErr)
{
  aErr err = aErrNone;
  unsigned long time;

  if ((err == aErrNone) 
      && (msTime > 0)) {
#if 0
    struct itimerval tv;
    int usec = (int)(msTime * 1000L);
 
    if (getitimer(ITIMER_REAL, &tv)) {
      printf("getitimer failed\n");
      err = aErrIO;
    } else {

      /* advance to the requested time */
      tv.it_value.tv_usec += usec % 1000000;
      tv.it_value.tv_sec += usec / 1000000;

      /* don't reload the timer */
      tv.it_interval.tv_usec = 0;
      tv.it_interval.tv_sec = 0;
 
      /* set the interval timer */
      if (setitimer(ITIMER_REAL, &tv, NULL)) {
        printf("setitimer failed\n");
        err = aErrIO;
      } else {
        struct sigaction new, save;
        sigset_t block;
      
        sigemptyset(&block);

	/* set up the action handler to go to our
	 * handler routine once */      
        aBZero(&new, sizeof(sigaction));
        new.sa_handler = alrm;
        new.sa_flags = SA_ONESHOT;
        sigemptyset(&new.sa_mask);

        /* now, block until the SIGALRM is received */
        if (sigaction(SIGALRM, &new, &save)) {
          printf("sigaction failed\n");
          err = aErrIO;
        } else {
          /* sigsuspend always returns with an error 
           * from the signal interrupt that unblocks it */
          sigsuspend(&block);
          save.sa_handler = SIG_DFL;
          sigaction(SIGALRM, &save, NULL);
        }
      }
    }
#else

    /* time is specified in microseconds, we want millis */
    time = msTime * 1000;

    /* usleep has a maximum of 1000000 so we break it up here */
    while ((err == aErrNone) && (time > 0)) {
      unsigned long inc;
      if (time > 999999)
        inc = 999999;
      else
        inc = time;
      time -= inc;

      /* usleep doesn't return an error on all unix platforms */
      usleep(inc);
    }

#endif
  }

  if (pErr != NULL)
    *pErr = err;

  return (aLIBRETURN)(err != aErrNone);

} /* aIO_MSSleep */

#endif /* aUNIX */
